#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
2300 - Module for testing object variables
"""

import datetime
import decimal
import unittest

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def __test_data(self, expected_int_value, expected_obj_value,
                    expected_array_value):
        int_value, object_value, array_value = self.cursor.fetchone()
        if object_value is not None:
            object_value = self.get_db_object_as_plain_object(object_value)
        if array_value is not None:
            array_value = array_value.aslist()
        self.assertEqual(int_value, expected_int_value)
        self.assertEqual(object_value, expected_obj_value)
        self.assertEqual(array_value, expected_array_value)

    def test_2300_bind_null_in(self):
        "2300 - test binding a null value (IN)"
        var = self.cursor.var(oracledb.DB_TYPE_OBJECT, typename="UDT_OBJECT")
        result = self.cursor.callfunc("pkg_TestBindObject.GetStringRep", str,
                                      (var,))
        self.assertEqual(result, "null")

    def test_2301_bind_object_in(self):
        "2301 - test binding an object (IN)"
        type_obj = self.connection.gettype("UDT_OBJECT")
        obj = type_obj.newobject()
        obj.NUMBERVALUE = 13
        obj.STRINGVALUE = "Test String"
        result = self.cursor.callfunc("pkg_TestBindObject.GetStringRep", str,
                                      (obj,))
        exp = "udt_Object(13, 'Test String', null, null, null, null, null)"
        self.assertEqual(result, exp)
        obj.NUMBERVALUE = None
        obj.STRINGVALUE = "Test With Dates"
        obj.DATEVALUE = datetime.datetime(2016, 2, 10)
        obj.TIMESTAMPVALUE = datetime.datetime(2016, 2, 10, 14, 13, 50)
        result = self.cursor.callfunc("pkg_TestBindObject.GetStringRep", str,
                                      (obj,))
        self.assertEqual(result,
                         "udt_Object(null, 'Test With Dates', null, " \
                         "to_date('2016-02-10', 'YYYY-MM-DD'), " \
                         "to_timestamp('2016-02-10 14:13:50', " \
                                 "'YYYY-MM-DD HH24:MI:SS'), " \
                         "null, null)")
        obj.DATEVALUE = None
        obj.TIMESTAMPVALUE = None
        sub_type_obj = self.connection.gettype("UDT_SUBOBJECT")
        sub_obj = sub_type_obj.newobject()
        sub_obj.SUBNUMBERVALUE = decimal.Decimal("18.25")
        sub_obj.SUBSTRINGVALUE = "Sub String"
        obj.SUBOBJECTVALUE = sub_obj
        result = self.cursor.callfunc("pkg_TestBindObject.GetStringRep", str,
                                      (obj,))
        self.assertEqual(result,
                         "udt_Object(null, 'Test With Dates', null, null, " \
                         "null, udt_SubObject(18.25, 'Sub String'), null)")

    def test_2302_copy_object(self):
        "2302 - test copying an object"
        type_obj = self.connection.gettype("UDT_OBJECT")
        obj = type_obj()
        obj.NUMBERVALUE = 5124
        obj.STRINGVALUE = "A test string"
        obj.DATEVALUE = datetime.datetime(2016, 2, 24)
        obj.TIMESTAMPVALUE = datetime.datetime(2016, 2, 24, 13, 39, 10)
        copied_obj = obj.copy()
        self.assertEqual(obj.NUMBERVALUE, copied_obj.NUMBERVALUE)
        self.assertEqual(obj.STRINGVALUE, copied_obj.STRINGVALUE)
        self.assertEqual(obj.DATEVALUE, copied_obj.DATEVALUE)
        self.assertEqual(obj.TIMESTAMPVALUE, copied_obj.TIMESTAMPVALUE)

    def test_2303_empty_collection_as_list(self):
        "2303 - test getting an empty collection as a list"
        type_obj = self.connection.gettype("UDT_ARRAY")
        obj = type_obj.newobject()
        self.assertEqual(obj.aslist(), [])

    def test_2304_fetch_data(self):
        "2304 - test fetching objects"
        self.cursor.execute("""
                select IntCol, ObjectCol, ArrayCol
                from TestObjects
                order by IntCol""")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ('OBJECTCOL', oracledb.DB_TYPE_OBJECT, None, None, None, None,
                    True),
            ('ARRAYCOL', oracledb.DB_TYPE_OBJECT, None, None, None, None, True)
        ]
        self.assertEqual(self.cursor.description, expected_value)
        expected_value = (
            1,
            'First row',
            'First     ',
            'N First Row',
            'N First   ',
            b'Raw Data 1',
            2,
            5,
            12.125,
            0.5,
            12.5,
            25.25,
            50.125,
            oracledb.Timestamp(2007, 3, 6, 0, 0, 0),
            oracledb.Timestamp(2008, 9, 12, 16, 40),
            oracledb.Timestamp(2009, 10, 13, 17, 50),
            oracledb.Timestamp(2010, 11, 14, 18, 55),
            'Short CLOB value',
            'Short NCLOB Value',
            b'Short BLOB value',
            (11, 'Sub object 1'),
            [(5, 'first element'), (6, 'second element')]
        )
        self.__test_data(1, expected_value, [5, 10, None, 20])
        self.__test_data(2, None, [3, None, 9, 12, 15])
        expected_value = (
            3,
            'Third row',
            'Third     ',
            'N Third Row',
            'N Third   ',
            b'Raw Data 3',
            4,
            10,
            6.5,
            0.75,
            43.25,
            86.5,
            192.125,
            oracledb.Timestamp(2007, 6, 21, 0, 0, 0),
            oracledb.Timestamp(2007, 12, 13, 7, 30, 45),
            oracledb.Timestamp(2017, 6, 21, 23, 18, 45),
            oracledb.Timestamp(2017, 7, 21, 8, 27, 13),
            'Another short CLOB value',
            'Another short NCLOB Value',
            b'Yet another short BLOB value',
            (13, 'Sub object 3'),
            [
                (10, 'element #1'),
                (20, 'element #2'),
                (30, 'element #3'),
                (40, 'element #4')
            ]
        )
        self.__test_data(3, expected_value, None)

    def test_2305_get_object_type(self):
        "2305 - test getting object type"
        type_obj = self.connection.gettype("UDT_OBJECT")
        self.assertEqual(type_obj.iscollection, False)
        self.assertEqual(type_obj.schema, self.connection.username.upper())
        self.assertEqual(type_obj.name, "UDT_OBJECT")
        sub_object_value_type = self.connection.gettype("UDT_SUBOBJECT")
        sub_object_array_type = self.connection.gettype("UDT_OBJECTARRAY")
        expected_attr_names = [
            "NUMBERVALUE",
            "STRINGVALUE",
            "FIXEDCHARVALUE",
            "NSTRINGVALUE",
            "NFIXEDCHARVALUE",
            "RAWVALUE",
            "INTVALUE",
            "SMALLINTVALUE",
            "REALVALUE",
            "DOUBLEPRECISIONVALUE",
            "FLOATVALUE",
            "BINARYFLOATVALUE",
            "BINARYDOUBLEVALUE",
            "DATEVALUE",
            "TIMESTAMPVALUE",
            "TIMESTAMPTZVALUE",
            "TIMESTAMPLTZVALUE",
            "CLOBVALUE",
            "NCLOBVALUE",
            "BLOBVALUE",
            "SUBOBJECTVALUE",
            "SUBOBJECTARRAY"
        ]
        actual_attr_names = [a.name for a in type_obj.attributes]
        self.assertEqual(actual_attr_names, expected_attr_names)
        expected_attr_types = [
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_VARCHAR,
            oracledb.DB_TYPE_CHAR,
            oracledb.DB_TYPE_NVARCHAR,
            oracledb.DB_TYPE_NCHAR,
            oracledb.DB_TYPE_RAW,
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_BINARY_FLOAT,
            oracledb.DB_TYPE_BINARY_DOUBLE,
            oracledb.DB_TYPE_DATE,
            oracledb.DB_TYPE_TIMESTAMP,
            oracledb.DB_TYPE_TIMESTAMP_TZ,
            oracledb.DB_TYPE_TIMESTAMP_LTZ,
            oracledb.DB_TYPE_CLOB,
            oracledb.DB_TYPE_NCLOB,
            oracledb.DB_TYPE_BLOB,
            sub_object_value_type,
            sub_object_array_type
        ]
        actual_attr_types = [a.type for a in type_obj.attributes]
        self.assertEqual(actual_attr_types, expected_attr_types)
        self.assertEqual(sub_object_array_type.iscollection, True)
        self.assertEqual(sub_object_array_type.attributes, [])

    def test_2306_object_type(self):
        "2306 - test object type data"
        self.cursor.execute("""
                select ObjectCol
                from TestObjects
                where ObjectCol is not null
                  and rownum <= 1""")
        obj, = self.cursor.fetchone()
        self.assertEqual(obj.type.schema, self.connection.username.upper())
        self.assertEqual(obj.type.name, "UDT_OBJECT")
        self.assertEqual(obj.type.attributes[0].name, "NUMBERVALUE")

    def test_2307_round_trip_object(self):
        "2307 - test inserting and then querying object with all data types"
        self.cursor.execute("truncate table TestClobs")
        self.cursor.execute("truncate table TestNClobs")
        self.cursor.execute("truncate table TestBlobs")
        self.cursor.execute("""
                insert into TestClobs
                (IntCol, ClobCol)
                values (1, 'A short CLOB')""")
        self.cursor.execute("""
                insert into TestNClobs
                (IntCol, NClobCol)
                values (1, 'A short NCLOB')""")
        self.cursor.execute("""
                insert into TestBlobs
                (IntCol, BlobCol)
                values (1, utl_raw.cast_to_raw('A short BLOB'))""")
        self.connection.commit()
        self.cursor.execute("select CLOBCol from TestClobs")
        clob, = self.cursor.fetchone()
        self.cursor.execute("select NCLOBCol from TestNClobs")
        nclob, = self.cursor.fetchone()
        self.cursor.execute("select BLOBCol from TestBlobs")
        blob, = self.cursor.fetchone()
        type_obj = self.connection.gettype("UDT_OBJECT")
        obj = type_obj.newobject()
        obj.NUMBERVALUE = 5
        obj.STRINGVALUE = "A string"
        obj.FIXEDCHARVALUE = "Fixed str"
        obj.NSTRINGVALUE = "A NCHAR string"
        obj.NFIXEDCHARVALUE = "Fixed N"
        obj.RAWVALUE = b"Raw Value"
        obj.INTVALUE = 27
        obj.SMALLINTVALUE = 13
        obj.REALVALUE = 184.875
        obj.DOUBLEPRECISIONVALUE = 1.375
        obj.FLOATVALUE = 23.75
        obj.DATEVALUE = datetime.date(2017, 5, 9)
        obj.TIMESTAMPVALUE = datetime.datetime(2017, 5, 9, 9, 41, 13)
        obj.TIMESTAMPTZVALUE = datetime.datetime(1986, 8, 2, 15, 27, 38)
        obj.TIMESTAMPLTZVALUE = datetime.datetime(1999, 11, 12, 23, 5, 2)
        obj.BINARYFLOATVALUE = 14.25
        obj.BINARYDOUBLEVALUE = 29.1625
        obj.CLOBVALUE = clob
        obj.NCLOBVALUE = nclob
        obj.BLOBVALUE = blob
        sub_type_obj = self.connection.gettype("UDT_SUBOBJECT")
        sub_obj = sub_type_obj.newobject()
        sub_obj.SUBNUMBERVALUE = 23
        sub_obj.SUBSTRINGVALUE = "Substring value"
        obj.SUBOBJECTVALUE = sub_obj
        self.cursor.execute("insert into TestObjects (IntCol, ObjectCol) " \
                            "values (4, :obj)", obj=obj)
        self.cursor.execute("""
                select IntCol, ObjectCol, ArrayCol
                from TestObjects
                where IntCol = 4""")
        expected_value = (
            5,
            'A string',
            'Fixed str ',
            'A NCHAR string',
            'Fixed N   ',
            b'Raw Value',
            27,
            13,
            184.875,
            1.375,
            23.75,
            14.25,
            29.1625,
            oracledb.Timestamp(2017, 5, 9, 0, 0, 0),
            oracledb.Timestamp(2017, 5, 9, 9, 41, 13),
            oracledb.Timestamp(1986, 8, 2, 15, 27, 38),
            oracledb.Timestamp(1999, 11, 12, 23, 5, 2),
            'A short CLOB',
            'A short NCLOB',
            b'A short BLOB',
            (23, 'Substring value'),
            None
        )
        self.__test_data(4, expected_value, None)
        obj.CLOBVALUE = "A short CLOB (modified)"
        obj.NCLOBVALUE = "A short NCLOB (modified)"
        obj.BLOBVALUE = "A short BLOB (modified)"
        self.cursor.execute("insert into TestObjects (IntCol, ObjectCol) " \
                            "values (5, :obj)", obj = obj)
        self.cursor.execute("""
                select IntCol, ObjectCol, ArrayCol
                from TestObjects
                where IntCol = 5""")
        expected_value = (
            5,
            'A string',
            'Fixed str ',
            'A NCHAR string',
            'Fixed N   ',
            b'Raw Value',
            27,
            13,
            184.875,
            1.375,
            23.75,
            14.25,
            29.1625,
            oracledb.Timestamp(2017, 5, 9, 0, 0, 0),
            oracledb.Timestamp(2017, 5, 9, 9, 41, 13),
            oracledb.Timestamp(1986, 8, 2, 15, 27, 38),
            oracledb.Timestamp(1999, 11, 12, 23, 5, 2),
            'A short CLOB (modified)',
            'A short NCLOB (modified)',
            b'A short BLOB (modified)',
            (23, 'Substring value'),
            None
        )
        self.__test_data(5, expected_value, None)
        self.connection.rollback()

    def test_2308_invalid_type_object(self):
        "2308 - test trying to find an object type that does not exist"
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2035:",
                               self.connection.gettype,
                               "A TYPE THAT DOES NOT EXIST")

    def test_2309_appending_wrong_object_type(self):
        "2309 - test appending an object of the wrong type to a collection"
        collection_obj_type = self.connection.gettype("UDT_OBJECTARRAY")
        collection_obj = collection_obj_type.newobject()
        array_obj_type = self.connection.gettype("UDT_ARRAY")
        array_obj = array_obj_type.newobject()
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2008:",
                               collection_obj.append, array_obj)

    def test_2310_referencing_sub_obj(self):
        "2310 - test that referencing a sub object affects the parent object"
        obj_type = self.connection.gettype("UDT_OBJECT")
        sub_obj_type = self.connection.gettype("UDT_SUBOBJECT")
        obj = obj_type.newobject()
        obj.SUBOBJECTVALUE = sub_obj_type.newobject()
        obj.SUBOBJECTVALUE.SUBNUMBERVALUE = 5
        obj.SUBOBJECTVALUE.SUBSTRINGVALUE = "Substring"
        self.assertEqual(obj.SUBOBJECTVALUE.SUBNUMBERVALUE, 5)
        self.assertEqual(obj.SUBOBJECTVALUE.SUBSTRINGVALUE, "Substring")

    def test_2311_access_sub_object_parent_object_destroyed(self):
        "2311 - test accessing sub object after parent object destroyed"
        obj_type = self.connection.gettype("UDT_OBJECT")
        sub_obj_type = self.connection.gettype("UDT_SUBOBJECT")
        array_type = self.connection.gettype("UDT_OBJECTARRAY")
        sub_obj1 = sub_obj_type.newobject()
        sub_obj1.SUBNUMBERVALUE = 2
        sub_obj1.SUBSTRINGVALUE = "AB"
        sub_obj2 = sub_obj_type.newobject()
        sub_obj2.SUBNUMBERVALUE = 3
        sub_obj2.SUBSTRINGVALUE = "CDE"
        obj = obj_type.newobject()
        obj.SUBOBJECTARRAY = array_type.newobject([sub_obj1, sub_obj2])
        sub_obj_array = obj.SUBOBJECTARRAY
        del obj
        self.assertEqual(self.get_db_object_as_plain_object(sub_obj_array),
                         [(2, "AB"), (3, "CDE")])

    def test_2312_setting_attr_wrong_object_type(self):
        "2312 - test assigning an object of wrong type to an object attribute"
        obj_type = self.connection.gettype("UDT_OBJECT")
        obj = obj_type.newobject()
        wrong_obj_type = self.connection.gettype("UDT_OBJECTARRAY")
        wrong_obj = wrong_obj_type.newobject()
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2008:", setattr,
                               obj, "SUBOBJECTVALUE", wrong_obj)

    def test_2313_setting_var_wrong_object_type(self):
        "2313 - test setting value of object variable to wrong object type"
        wrong_obj_type = self.connection.gettype("UDT_OBJECTARRAY")
        wrong_obj = wrong_obj_type.newobject()
        var = self.cursor.var(oracledb.DB_TYPE_OBJECT, typename="UDT_OBJECT")
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2008:",
                               var.setvalue, 0, wrong_obj)

    def test_2314_string_format(self):
        "2314 - test object string format"
        obj_type = self.connection.gettype("UDT_OBJECT")
        user = test_env.get_main_user()
        self.assertEqual(str(obj_type),
                "<oracledb.DbObjectType %s.UDT_OBJECT>" % user.upper())
        self.assertEqual(str(obj_type.attributes[0]),
                         "<oracledb.DbObjectAttr NUMBERVALUE>")

    def test_2315_trim_collection_list(self):
        "2315 - test Trim number of elements from collection"
        sub_obj_type = self.connection.gettype("UDT_SUBOBJECT")
        array_type = self.connection.gettype("UDT_OBJECTARRAY")
        data = [(1, "AB"), (2, "CDE"), (3, "FGH"), (4, "IJK"), (5, "LMN")]
        array_obj = array_type()
        for num_val, str_val in data:
            subObj = sub_obj_type()
            subObj.SUBNUMBERVALUE = num_val
            subObj.SUBSTRINGVALUE = str_val
            array_obj.append(subObj)
        self.assertEqual(self.get_db_object_as_plain_object(array_obj), data)
        array_obj.trim(2)
        self.assertEqual(self.get_db_object_as_plain_object(array_obj),
                         data[:3])
        array_obj.trim(1)
        self.assertEqual(self.get_db_object_as_plain_object(array_obj),
                         data[:2])
        array_obj.trim(0)
        self.assertEqual(self.get_db_object_as_plain_object(array_obj),
                         data[:2])
        array_obj.trim(2)
        self.assertEqual(self.get_db_object_as_plain_object(array_obj), [])

    def test_2316_sql_type_metadata(self):
        "2316 - test the metadata of a SQL type"
        user = test_env.get_main_user()
        typ = self.connection.gettype("UDT_OBJECTARRAY")
        self.assertEqual(typ.schema, user.upper())
        self.assertEqual(typ.name, "UDT_OBJECTARRAY")
        self.assertEqual(typ.package_name, None)
        self.assertEqual(typ.element_type.schema, user.upper())
        self.assertEqual(typ.element_type.name, "UDT_SUBOBJECT")
        self.assertEqual(typ.element_type.package_name, None)

    def test_2317_plsql_type_metadata(self):
        "2317 - test the metadata of a PL/SQL type"
        user = test_env.get_main_user()
        typ = self.connection.gettype("PKG_TESTSTRINGARRAYS.UDT_STRINGLIST")
        self.assertEqual(typ.schema, user.upper())
        self.assertEqual(typ.name, "UDT_STRINGLIST")
        self.assertEqual(typ.package_name, "PKG_TESTSTRINGARRAYS")
        self.assertEqual(typ.element_type, oracledb.DB_TYPE_VARCHAR)

    def test_2318_negative_create_object_var_no_type_name(self):
        "2318 - test creating an object variable without a type name"
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2037:",
                               self.cursor.var, oracledb.DB_TYPE_OBJECT)

    def test_2319_collection_as_dictionary(self):
        "2319 - test getting an empty collection as a dictionary"
        type_obj = self.connection.gettype("UDT_ARRAY")
        obj = type_obj.newobject()
        self.assertEqual(obj.asdict(), {})

    def test_2320_exists_in_collection(self):
        "2320 - test if an element exists in a collection"
        array_type = self.connection.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertEqual(array_obj.exists(0), False)
        array_obj.append(40)
        self.assertEqual(array_obj.exists(0), True)
        self.assertEqual(array_obj.exists(1), False)

    def test_2321_first_and_last_indexes(self):
        "2321 - test first and last methods"
        array_type = self.connection.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertEqual(array_obj.first(), None)
        self.assertEqual(array_obj.last(), None)
        for i in range(7):
            array_obj.append(i)
        self.assertEqual(array_obj.first(), 0)
        self.assertEqual(array_obj.last(), 6)

    def test_2322_size(self):
        "2322 - test getting the size of a collections"
        array_type = self.connection.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertEqual(array_obj.size(), 0)
        for i in range(5):
            array_obj.append(i)
        self.assertEqual(array_obj.size(), 5)

    def test_2323_prev_next(self):
        "2323 - test prev and next methods"
        array_type = self.connection.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertEqual(array_obj.prev(0), None)
        self.assertEqual(array_obj.next(0), None)
        for i in range(2):
            array_obj.append(i)
        self.assertEqual(array_obj.prev(0), None)
        self.assertEqual(array_obj.prev(1), 0)
        self.assertEqual(array_obj.next(0), 1)
        self.assertEqual(array_obj.next(1), None)

    def test_2324_set_get_element(self):
        "2324 - test setting and getting elements from a collection"
        array_type = self.connection.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2038:",
                               array_obj.getelement, 0)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2039:",
                               array_obj.setelement, 0, 7)
        array_obj.append(7)
        self.assertEqual(array_obj.getelement(0), 7)
        array_obj.setelement(0, 10)
        self.assertEqual(array_obj.getelement(0), 10)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2039:",
                               array_obj.setelement, 3, 4)

    def test_2325_add_too_many_elements_to_collection(self):
        "2325 - test appending too many elements to a collection"
        array_type = self.connection.gettype("UDT_ARRAY")
        numbers = [i for i in range(11)]
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2039:",
                               array_type.newobject, numbers)

        array_obj = array_type.newobject()
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2039:",
                               array_obj.extend, numbers)

        array_obj = array_type.newobject()
        for elem in numbers[:10]:
            array_obj.append(elem)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2039:",
                               array_obj.append, numbers[10])

    def test_2326_test_unconstrained_table(self):
        "2326 - test appending elements to an unconstrained table"
        data = [1, 3, 6, 10, 15, 21]
        typ = self.connection.gettype("UDT_UNCONSTRAINEDTABLE")
        obj = typ.newobject(data)
        self.cursor.execute("select :1 from dual", [obj])
        output_obj, = self.cursor.fetchone()
        self.assertEqual(output_obj.aslist(), data)

    def test_2327_large_collection(self):
        "2327 - test collection with thousands of entries"
        typ = self.connection.gettype("PKG_TESTNUMBERARRAYS.UDT_NUMBERLIST")
        obj = typ.newobject()
        obj.setelement(1, 1)
        running_total = 1
        for i in range(1, 35000):
            running_total += (i + 1)
            obj.append(running_total)
        result = self.cursor.callfunc("pkg_TestNumberArrays.TestInArrays", int,
                                      (2327, obj))
        self.assertEqual(result, 7146445847327)

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't have any unknown types currently")
    def test_2328_unknown_type_attribute(self):
        "2328 - test object with unknown type in one of its attributes"
        typ = self.connection.gettype("UDT_UNKNOWNATTRIBUTETYPE")
        self.assertEqual(typ.attributes[1].type, oracledb.DB_TYPE_UNKNOWN)

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't have any unknown types currently")
    def test_2329_unknown_type_element(self):
        "2329 - test object with unknown type as the element type"
        typ = self.connection.gettype("UDT_UNKNOWNELEMENTTYPE")
        self.assertEqual(typ.element_type, oracledb.DB_TYPE_UNKNOWN)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# var.pyx
#
# Cython file defining the base Variable implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseVarImpl:

    cdef int _bind(self, object conn, BaseCursorImpl cursor,
                   uint32_t num_execs, object name, uint32_t pos) except -1:
        """
        Binds a variable to the cursor.
        """
        raise NotImplementedError()


    cdef int _check_and_set_scalar_value(self, uint32_t pos, object value,
                                         bint* was_set) except -1:
        """
        Sets a scalar value in the variable at the given position, but first
        checks the type of Python value to see if it is acceptable. The value
        may be modified by the in converter (if one has been set) or adjusted
        to be acceptable (for some cases). If the was_set pointer is NULL, an
        exception is raised when the Python value is found to be unacceptable;
        otherwise, the flag is cleared if the Python value is unacceptable.
        """
        cdef uint32_t size

        # call in converter, if applicable
        if self.inconverter is not None:
            value = self.inconverter(value)

        # check the value and verify it is acceptable
        value = self._conn_impl._check_value(self.dbtype, self.objtype, value,
                                             was_set)
        if was_set != NULL and not was_set[0]:
            return 0

        # resize variable, if applicable
        if value is not None and self.dbtype.default_size != 0:
            size = <uint32_t> len(value)
            if size > self.size:
                self._resize(size)

        # set value
        self._set_scalar_value(pos, value)
        self._is_value_set = True

    cdef int _check_and_set_value(self, uint32_t pos, object value,
                                  bint* was_set) except -1:
        """
        Sets the value in the variable at the given position, but first checks
        the type of Python value to see if it is acceptable.
        """
        cdef:
            uint32_t i, num_elements_in_array
            object element_value

        # scalar variables can be checked directly
        if not self.is_array:
            return self._check_and_set_scalar_value(pos, value, was_set)

        # array variables must have a list supplied to them
        if not isinstance(value, list):
            if was_set != NULL:
                was_set[0] = False
                return 0
            errors._raise_err(errors.ERR_EXPECTING_LIST_FOR_ARRAY_VAR)

        # the size of the array must be sufficient to hold all of the
        # elements
        num_elements_in_array = len(<list> value)
        if num_elements_in_array > self.num_elements:
            if was_set != NULL:
                was_set[0] = False
                return 0
            errors._raise_err(errors.ERR_INCORRECT_VAR_ARRAYSIZE,
                              var_arraysize=self.num_elements,
                              required_arraysize=num_elements_in_array)

        # check and set each of the element's values
        for i, element_value in enumerate(<list> value):
            self._check_and_set_scalar_value(i, element_value, was_set)
            if was_set != NULL and not was_set[0]:
                return 0
        self._set_num_elements_in_array(num_elements_in_array)

    cdef int _finalize_init(self) except -1:
        """
        Internal method that finalizes initialization of the variable.
        """
        if self.dbtype.default_size > 0:
            if self.size == 0:
                self.size = self.dbtype.default_size
            self.buffer_size = self.size * self.dbtype._buffer_size_factor
        else:
            self.buffer_size = self.dbtype._buffer_size_factor
        if self.num_elements == 0:
            self.num_elements = 1

    cdef list _get_array_value(self):
        """
        Internal method to return the value of the array.
        """
        raise NotImplementedError()

    cdef object _get_scalar_value(self, uint32_t pos):
        """
        Internal method to return the value of the variable at the given
        position.
        """
        raise NotImplementedError()

    cdef int _on_reset_bind(self, uint32_t num_rows) except -1:
        """
        Called when the bind variable is being reset, just prior to performing
        a bind operation.
        """
        if self.num_elements < num_rows:
            self.num_elements = num_rows
            self._finalize_init()

    cdef int _resize(self, uint32_t new_size) except -1:
        """
        Resize the variable to the new size provided.
        """
        self.size = new_size
        self.buffer_size = new_size * self.dbtype._buffer_size_factor

    cdef int _set_scalar_value(self, uint32_t pos, object value) except -1:
        """
        Set the value of the variable at the given position. At this point it
        is assumed that all checks have been performed!
        """
        raise NotImplementedError()

    cdef int _set_num_elements_in_array(self, uint32_t num_elements) except -1:
        """
        Sets the number of elements in the array.
        """
        self.num_elements_in_array = num_elements

    cdef int _set_type_info_from_type(self, object typ) except -1:
        """
        Sets the type and size of the variable given a Python type.
        """
        cdef ApiType apitype
        if isinstance(typ, DbType):
            self.dbtype = typ
        elif isinstance(typ, ApiType):
            apitype = typ
            self.dbtype = apitype.dbtypes[0]
        elif isinstance(typ, PY_TYPE_DB_OBJECT_TYPE):
            self.dbtype = DB_TYPE_OBJECT
            self.objtype = typ._impl
        elif not isinstance(typ, type):
            errors._raise_err(errors.ERR_EXPECTING_TYPE)
        elif typ is int:
            self.dbtype = DB_TYPE_NUMBER
            self._preferred_num_type = NUM_TYPE_INT
        elif typ is float:
            self.dbtype = DB_TYPE_NUMBER
            self._preferred_num_type = NUM_TYPE_FLOAT
        elif typ is str:
            self.dbtype = DB_TYPE_VARCHAR
        elif typ is bytes:
            self.dbtype = DB_TYPE_RAW
        elif typ is PY_TYPE_DECIMAL:
            self.dbtype = DB_TYPE_NUMBER
            self._preferred_num_type = NUM_TYPE_DECIMAL
        elif typ is PY_TYPE_BOOL:
            self.dbtype = DB_TYPE_BOOLEAN
        elif typ is PY_TYPE_DATE:
            self.dbtype = DB_TYPE_DATE
        elif typ is PY_TYPE_DATETIME:
            self.dbtype = DB_TYPE_TIMESTAMP
        elif typ is PY_TYPE_TIMEDELTA:
            self.dbtype = DB_TYPE_INTERVAL_DS
        else:
            errors._raise_err(errors.ERR_PYTHON_TYPE_NOT_SUPPORTED, typ=typ)

    cdef int _set_type_info_from_value(self, object value,
                                       bint is_plsql) except -1:
        """
        Sets the type and size of the variable given a Python value. This
        method is called once for scalars and once per element in a list for
        array values. If a different type is detected an error is raised.
        """
        cdef:
            int preferred_num_type = NUM_TYPE_FLOAT
            BaseDbObjectTypeImpl objtype = None
            DbType dbtype = None
            uint32_t size = 0
        if value is None:
            dbtype = DB_TYPE_VARCHAR
            size = 1
        elif isinstance(value, PY_TYPE_BOOL):
            dbtype = DB_TYPE_BOOLEAN if is_plsql else DB_TYPE_BINARY_INTEGER
        elif isinstance(value, str):
            size = <uint32_t> len(value)
            dbtype = DB_TYPE_VARCHAR
        elif isinstance(value, bytes):
            size = <uint32_t> len(value)
            dbtype = DB_TYPE_RAW
        elif isinstance(value, int):
            dbtype = DB_TYPE_NUMBER
            preferred_num_type = NUM_TYPE_INT
        elif isinstance(value, float):
            dbtype = DB_TYPE_NUMBER
            preferred_num_type = NUM_TYPE_FLOAT
        elif isinstance(value, PY_TYPE_DECIMAL):
            dbtype = DB_TYPE_NUMBER
            preferred_num_type = NUM_TYPE_DECIMAL
        elif isinstance(value, (PY_TYPE_DATE, PY_TYPE_DATETIME)):
            dbtype = DB_TYPE_DATE
        elif isinstance(value, PY_TYPE_TIMEDELTA):
            dbtype = DB_TYPE_INTERVAL_DS
        elif isinstance(value, PY_TYPE_DB_OBJECT):
            dbtype = DB_TYPE_OBJECT
            objtype = value.type._impl
        elif isinstance(value, PY_TYPE_LOB):
            dbtype = value.type
        elif isinstance(value, PY_TYPE_CURSOR):
            dbtype = DB_TYPE_CURSOR
        else:
            errors._raise_err(errors.ERR_PYTHON_VALUE_NOT_SUPPORTED,
                              type_name=type(value).__name__)
        if self.dbtype is None:
            self.dbtype = dbtype
            self.objtype = objtype
            self._preferred_num_type = preferred_num_type
        elif dbtype is not self.dbtype or objtype is not self.objtype:
            errors._raise_err(errors.ERR_MIXED_ELEMENT_TYPES, element=value)
        if size > self.size:
            self.size = size

    def get_description(self):
        """
        Return a 7-tuple containing information about the variable: (name,
        type, display_size, internal_size, precision, scale, null_ok).
        """
        cdef:
            uint32_t display_size = 0, size_in_bytes = 0
            FetchInfo fetch_info = self._fetch_info
            DbType dbtype = fetch_info._dbtype
            object precision, scale
        if fetch_info._size > 0:
            display_size = fetch_info._size
            size_in_bytes = fetch_info._buffer_size
        elif dbtype is DB_TYPE_DATE \
                or dbtype is DB_TYPE_TIMESTAMP \
                or dbtype is DB_TYPE_TIMESTAMP_LTZ \
                or dbtype is DB_TYPE_TIMESTAMP_TZ:
            display_size = 23
        elif dbtype is DB_TYPE_BINARY_FLOAT \
                or dbtype is DB_TYPE_BINARY_DOUBLE \
                or dbtype is DB_TYPE_BINARY_INTEGER \
                or dbtype is DB_TYPE_NUMBER:
            if fetch_info._precision:
                display_size = fetch_info._precision + 1
                if fetch_info._scale > 0:
                    display_size += fetch_info._scale + 1
            else:
                display_size = 127
        if fetch_info._precision or fetch_info._scale:
            precision = fetch_info._precision
            scale = fetch_info._scale
        else:
            scale = precision = None
        return (fetch_info._name, dbtype, display_size or None,
                size_in_bytes or None, precision, scale,
                fetch_info._nulls_allowed)

    def get_all_values(self):
        """
        Internal method for returning an array of all of the values stored in
        the variable.
        """
        cdef uint32_t i
        if self.is_array:
            return self._get_array_value()
        return [self._get_scalar_value(i) for i in range(self.num_elements)]

    def get_value(self, uint32_t pos):
        """
        Internal method for getting the value of a variable.
        """
        if self.is_array:
            return self._get_array_value()
        if pos >= self.num_elements:
            raise IndexError("position out of range")
        return self._get_scalar_value(pos)

    def set_value(self, uint32_t pos, object value):
        """
        Internal method for setting a variable's value at the specified
        position.
        """
        if self.is_array:
            if pos > 0:
                errors._raise_err(errors.ERR_ARRAYS_OF_ARRAYS)
        elif pos >= self.num_elements:
            raise IndexError("position out of range")
        self._check_and_set_value(pos, value, NULL)



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3800 - Module for testing the input and output type handlers.
"""

import json
import unittest

import oracledb
import test_env

class Building(object):

    def __init__(self, building_id, description, num_floors):
        self.building_id = building_id
        self.description = description
        self.num_floors = num_floors

    def __repr__(self):
        return "<Building %s: %s>" % (self.building_id, self.description)

    def __eq__(self, other):
        if isinstance(other, Building):
            return other.building_id == self.building_id \
                    and other.description == self.description \
                    and other.num_floors == self.num_floors
        return NotImplemented

    def to_json(self):
        return json.dumps(self.__dict__)

    @classmethod
    def from_json(cls, value):
        result = json.loads(value)
        return cls(**result)


class TestCase(test_env.BaseTestCase):

    def building_in_converter(self, value):
        return value.to_json()

    def input_type_handler(self, cursor, value, num_elements):
        if isinstance(value, Building):
            return cursor.var(oracledb.STRING, arraysize=num_elements,
                              inconverter=self.building_in_converter)

    def output_type_handler(self, cursor, name, default_type, size, precision,
                            scale):
        if default_type == oracledb.STRING:
            return cursor.var(default_type, arraysize=cursor.arraysize,
                              outconverter=Building.from_json)

    def test_3800(self):
        "3800 - binding unsupported python object without input type handler"
        self.cursor.execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        building = Building(1, "The First Building", 5)
        self.assertRaisesRegex(oracledb.NotSupportedError, "^DPY-3002:",
                               self.cursor.execute, sql,
                               (building.building_id, building))

    def test_3801(self):
        "3801 - not callable input type handler"
        self.cursor.execute("truncate table TestTempTable")
        building = Building(1, "The First Building", 5)
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.cursor.inputtypehandler = 5
        self.assertRaises(TypeError, self.cursor.execute, sql,
                          (building.building_id, building))

    def test_3802(self):
        "3802 - binding unsupported python object with input type handler"
        self.cursor.execute("truncate table TestTempTable")
        building = Building(1, "The First Building", 5)
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.cursor.inputtypehandler = self.input_type_handler
        self.cursor.execute(sql, (building.building_id, building))
        self.connection.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(self.cursor.fetchall(),
                         [(building.building_id, building.to_json())])

    def test_3803(self):
        "3803 - input type handler and output type handler on cursor level"
        self.cursor.execute("truncate table TestTempTable")
        building_one = Building(1, "The First Building", 5)
        building_two = Building(2, "The Second Building", 87)
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        cursor_one = self.connection.cursor()
        cursor_two = self.connection.cursor()
        cursor_one.inputtypehandler = self.input_type_handler
        cursor_one.execute(sql, (building_one.building_id, building_one))
        self.connection.commit()
        cursor_one.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(cursor_one.fetchall(),
                         [(building_one.building_id, building_one.to_json())])
        self.assertRaisesRegex(oracledb.NotSupportedError, "^DPY-3002:",
                               cursor_two.execute, sql,
                               (building_two.building_id, building_two))
        cursor_two.outputtypehandler = self.output_type_handler
        cursor_two.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(cursor_two.fetchall(),
                         [(building_one.building_id, building_one)])

    def test_3804(self):
        "3804 - input type handler and output type handler on connection level"
        self.cursor.execute("truncate table TestTempTable")
        building_one = Building(1, "The First Building", 5)
        building_two = Building(2, "The Second Building", 87)
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        connection = test_env.get_connection()
        connection.inputtypehandler = self.input_type_handler
        cursor_one = connection.cursor()
        cursor_two = connection.cursor()
        cursor_one.execute(sql, (building_one.building_id, building_one))
        cursor_two.execute(sql, (building_two.building_id, building_two))
        connection.commit()
        expected_data = [
            (building_one.building_id, building_one),
            (building_two.building_id, building_two)
        ]
        connection.outputtypehandler = self.output_type_handler
        cursor_one.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(cursor_one.fetchall(), expected_data)
        cursor_two.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(cursor_two.fetchall(), expected_data)
        other_cursor = self.connection.cursor()
        self.assertRaisesRegex(oracledb.NotSupportedError, "^DPY-3002:",
                               other_cursor.execute, sql,
                               (building_one.building_id, building_one))

    def test_3805(self):
        "3805 - output type handler with outconvert and null values"
        self.cursor.execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        data_to_insert = [
            (1, "String 1"),
            (2, None),
            (3, "String 2")
        ]
        self.cursor.executemany(sql, data_to_insert)
        self.connection.commit()
        def converter(value):
            return "CONVERTED"
        def output_type_handler(cursor, name, default_type, size, precision,
                                scale):
            if default_type is oracledb.DB_TYPE_VARCHAR:
                return cursor.var(str, outconverter=converter,
                                  arraysize=cursor.arraysize)
        self.cursor.outputtypehandler = output_type_handler
        self.cursor.execute("""
                select IntCol, StringCol1
                from TestTempTable
                order by IntCol""")
        expected_data = [
            (1, "CONVERTED"),
            (2, None),
            (3, "CONVERTED")
        ]
        self.assertEqual(self.cursor.fetchall(), expected_data)

    @unittest.skipUnless(test_env.get_server_version() >= (21, 0),
                         "unsupported server")
    def test_3806(self):
        "3806 - output type handler for fetching 21c JSON"
        self.cursor.execute("truncate table TestJson")
        insert_sql = "insert into TestJson values (:1, :2)"
        json_data = [
            dict(name="John", city="Delhi"),
            dict(name="George", city="Bangalore"),
            dict(name="Sam", city="Mumbai")
        ]
        data_to_insert = list(enumerate(json_data))
        json_as_string = self.connection.thin or \
                test_env.get_client_version() < (21, 0)
        if json_as_string:
            # insert data as JSON string
            self.cursor.executemany(insert_sql,
                    [(i, json.dumps(j)) for i, j in data_to_insert])
        else:
            # take advantage of direct binding
            self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON)
            self.cursor.executemany(insert_sql, data_to_insert)
        def output_type_handler(cursor, name, default_type, size, precision,
                                scale):
            # fetch 21c JSON datatype when using python-oracledb thin mode
            if default_type == oracledb.DB_TYPE_JSON:
                return cursor.var(str, arraysize=cursor.arraysize,
                                  outconverter=json.loads)
            # if using Oracle Client version < 21, then database returns BLOB
            # data type instead of JSON data type
            elif default_type == oracledb.DB_TYPE_BLOB:
                return cursor.var(default_type, arraysize=cursor.arraysize,
                                  outconverter=lambda v: json.loads(v.read()))
        if json_as_string:
            self.cursor.outputtypehandler = output_type_handler
        self.cursor.execute("select * from TestJson")
        self.assertEqual(self.cursor.fetchall(), data_to_insert)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
1900 - Module for testing LOB (CLOB and BLOB) variables
"""

import pickle

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def __get_temp_lobs(self, sid):
        cursor = self.connection.cursor()
        cursor.execute("""
                select cache_lobs + nocache_lobs + abstract_lobs
                from v$temporary_lobs
                where sid = :sid""", sid = sid)
        row = cursor.fetchone()
        if row is None:
            return 0
        return int(row[0])

    def __perform_test(self, lob_type, input_type):
        long_string = ""
        db_type = getattr(oracledb, "DB_TYPE_" + lob_type)
        self.cursor.execute(f"truncate table Test{lob_type}s")
        for i in range(0, 11):
            if i > 0:
                char = chr(ord('A') + i - 1)
                long_string += char * 25000
            elif input_type is not db_type:
                continue
            self.cursor.setinputsizes(long_string=input_type)
            if lob_type == "BLOB":
                bind_value = long_string.encode()
            else:
                bind_value = long_string
            self.cursor.execute(f"""
                    insert into Test{lob_type}s (
                        IntCol,
                        {lob_type}Col
                    ) values (
                        :integer_value,
                        :long_string
                    )""",
                    integer_value=i,
                    long_string=bind_value)
        self.connection.commit()
        self.cursor.execute(f"""
                select
                    IntCol,
                    {lob_type}Col
                from Test{lob_type}s
                order by IntCol""")
        self.__validate_query(self.cursor, lob_type)

    def __test_bind_ordering(self, lob_type):
        main_col = "A" * 32768
        extra_col_1 = "B" * 65536
        extra_col_2 = "C" * 131072
        if lob_type == "BLOB":
            main_col = main_col.encode()
            extra_col_1 = extra_col_1.encode()
            extra_col_2 = extra_col_2.encode()
        self.connection.stmtcachesize = 0
        self.cursor.execute(f"truncate table Test{lob_type}s")
        data = (1, main_col, 8, extra_col_1, 15, extra_col_2)
        self.cursor.execute(f"""
                insert into Test{lob_type}s (
                    IntCol,
                    {lob_type}Col,
                    ExtraNumCol1,
                    Extra{lob_type}Col1,
                    ExtraNumCol2,
                    Extra{lob_type}Col2
                ) values (:1, :2, :3, :4, :5, :6)""",
                data)
        with test_env.FetchLobsContextManager(False):
            self.cursor.execute(f"select * from Test{lob_type}s")
            fetched_data = self.cursor.fetchone()
            self.assertEqual(fetched_data, data)

    def __test_fetch_lobs_direct(self, lob_type):
        self.cursor.execute(f"truncate table Test{lob_type}s")
        data = []
        long_string = ""
        for i in range(1, 11):
            if i > 0:
                char = chr(ord('A') + i - 1)
                long_string += char * 25000
            if lob_type == "BLOB":
                data.append((i, long_string.encode()))
            else:
                data.append((i, long_string))
        self.cursor.executemany(f"""
                    insert into Test{lob_type}s (
                        IntCol,
                        {lob_type}Col
                    ) values (
                        :1,
                        :2
                    )""", data)
        with test_env.FetchLobsContextManager(False):
            self.cursor.execute(f"""
                    select
                        IntCol,
                        {lob_type}Col
                    from Test{lob_type}s
                    order by IntCol""")
            self.assertEqual(data, self.cursor.fetchall())

    def __test_lob_operations(self, lob_type):
        self.cursor.execute(f"truncate table Test{lob_type}s")
        self.cursor.setinputsizes(long_string=getattr(oracledb, lob_type))
        long_string = "X" * 75000
        write_value = "TEST"
        if lob_type == "BLOB":
            long_string = long_string.encode("ascii")
            write_value = write_value.encode("ascii")
        self.cursor.execute(f"""
                insert into Test{lob_type}s (
                    IntCol,
                    {lob_type}Col
                ) values (
                    :integer_value,
                    :long_string
                )""",
                integer_value=1,
                long_string=long_string)
        self.cursor.execute(f"""
                select {lob_type}Col
                from Test{lob_type}s
                where IntCol = 1""")
        lob, = self.cursor.fetchone()
        self.assertEqual(lob.isopen(), False)
        lob.open()
        self.assertEqual(lob.isopen(), True)
        lob.close()
        self.assertEqual(lob.isopen(), False)
        self.assertEqual(lob.size(), 75000)
        lob.write(write_value, 75001)
        self.assertEqual(lob.size(), 75000 + len(write_value))
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2030:",
                               lob.read, 0)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2030:",
                               lob.read, -25)
        self.assertEqual(lob.read(), long_string + write_value)
        lob.write(write_value, 1)
        self.assertEqual(lob.read(),
                         write_value + long_string[4:] + write_value)
        lob.trim(25000)
        self.assertEqual(lob.size(), 25000)
        lob.trim(newSize=10000)
        self.assertEqual(lob.size(), 10000)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2014:",
                               lob.trim, new_size=50, newSize=60)
        lob.trim()
        self.assertEqual(lob.size(), 0)
        self.assertIsInstance(lob.getchunksize(), int)

    def __test_pickle(self, lob_type):
        value = "A test string value for pickling"
        if lob_type == "BLOB":
            value = value.encode("ascii")
        db_type = getattr(oracledb, "DB_TYPE_" + lob_type)
        lob = self.connection.createlob(db_type)
        lob.write(value)
        pickled_data = pickle.dumps(lob)
        unpickled_value = pickle.loads(pickled_data)
        self.assertEqual(unpickled_value, value)

    def __test_temporary_lob(self, lob_type):
        self.cursor.execute(f"truncate table Test{lob_type}s")
        value = "A test string value"
        if lob_type == "BLOB":
            value = value.encode("ascii")
        db_type = getattr(oracledb, "DB_TYPE_" + lob_type)
        lob = self.connection.createlob(db_type)
        lob.write(value)
        self.cursor.execute("""
                insert into Test%ss (IntCol, %sCol)
                values (:int_val, :lob_val)""" % (lob_type, lob_type),
                int_val=1,
                lob_val=lob)
        self.cursor.execute("select %sCol from Test%ss" % (lob_type, lob_type))
        lob, = self.cursor.fetchone()
        self.assertEqual(lob.read(), value)

    def __validate_query(self, rows, lob_type):
        long_string = ""
        db_type = getattr(oracledb, "DB_TYPE_" + lob_type)
        for row in rows:
            integer_value, lob = row
            self.assertEqual(lob.type, db_type)
            if integer_value == 0:
                self.assertEqual(lob.size(), 0)
                expected_value = ""
                if lob_type == "BLOB":
                    expected_value = expected_value.encode()
                self.assertEqual(lob.read(), expected_value)
            else:
                char = chr(ord('A') + integer_value - 1)
                prev_char = chr(ord('A') + integer_value - 2)
                long_string += char * 25000
                if lob_type == "BLOB":
                    expected_value = long_string.encode("ascii")
                    char = char.encode("ascii")
                    prev_char = prev_char.encode("ascii")
                else:
                    expected_value = long_string
                self.assertEqual(lob.size(), len(expected_value))
                self.assertEqual(lob.read(), expected_value)
                if lob_type == "CLOB":
                    self.assertEqual(str(lob), expected_value)
                self.assertEqual(lob.read(len(expected_value)), char)
            if integer_value > 1:
                offset = (integer_value - 1) * 25000 - 4
                string = prev_char * 5 + char * 5
                self.assertEqual(lob.read(offset, 10), string)

    def test_1900_bind_lob_value(self):
        "1900 - test binding a LOB value directly"
        self.cursor.execute("truncate table TestCLOBs")
        self.cursor.execute("""
                insert into TestCLOBs
                (IntCol, ClobCol)
                values (1, 'Short value')""")
        self.cursor.execute("select ClobCol from TestCLOBs")
        lob, = self.cursor.fetchone()
        self.cursor.execute("""
                insert into TestCLOBs
                (IntCol, ClobCol)
                values (2, :value)""",
                value=lob)

    def test_1901_blob_cursor_description(self):
        "1901 - test cursor description is accurate for BLOBs"
        self.cursor.execute("select IntCol, BlobCol from TestBLOBs")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, 0),
            ('BLOBCOL', oracledb.DB_TYPE_BLOB, None, None, None, None, 0)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1902_blob_direct(self):
        "1902 - test binding and fetching BLOB data (directly)"
        self.__perform_test("BLOB", oracledb.DB_TYPE_BLOB)

    def test_1903_blob_indirect(self):
        "1903 - test binding and fetching BLOB data (indirectly)"
        self.__perform_test("BLOB", oracledb.DB_TYPE_LONG_RAW)

    def test_1904_blob_operations(self):
        "1904 - test operations on BLOBs"
        self.__test_lob_operations("BLOB")

    def test_1905_clob_cursor_description(self):
        "1905 - test cursor description is accurate for CLOBs"
        self.cursor.execute("select IntCol, ClobCol from TestCLOBs")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ('CLOBCOL', oracledb.DB_TYPE_CLOB, None, None, None, None, False)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1906_clob_direct(self):
        "1906 - test binding and fetching CLOB data (directly)"
        self.__perform_test("CLOB", oracledb.DB_TYPE_CLOB)

    def test_1907_clob_indirect(self):
        "1907 - test binding and fetching CLOB data (indirectly)"
        self.__perform_test("CLOB", oracledb.DB_TYPE_LONG)

    def test_1908_clob_operations(self):
        "1908 - test operations on CLOBs"
        self.__test_lob_operations("CLOB")

    def test_1909_create_temp_blob(self):
        "1909 - test creating a temporary BLOB"
        self.__test_temporary_lob("BLOB")

    def test_1910_create_temp_clob(self):
        "1910 - test creating a temporary CLOB"
        self.__test_temporary_lob("CLOB")

    def test_1911_create_temp_nclob(self):
        "1911 - test creating a temporary NCLOB"
        self.__test_temporary_lob("NCLOB")

    def test_1912_multiple_fetch(self):
        "1912 - test retrieving data from a CLOB after multiple fetches"
        self.cursor.arraysize = 1
        self.cursor.execute("select * from TestCLOBS")
        rows = self.cursor.fetchall()
        self.__validate_query(rows, "CLOB")

    def test_1913_nclob_cursor_description(self):
        "1913 - test cursor description is accurate for NCLOBs"
        self.cursor.execute("select IntCol, NClobCol from TestNCLOBs")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, 0),
            ('NCLOBCOL', oracledb.DB_TYPE_NCLOB, None, None, None, None, 0)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1914_nclob_direct(self):
        "1914 - test binding and fetching NCLOB data (directly)"
        self.__perform_test("NCLOB", oracledb.DB_TYPE_NCLOB)

    def test_1915_nclob_non_ascii_chars(self):
        "1915 - test binding and fetching NCLOB data (with non-ASCII chars)"
        value = "\u03b4\u4e2a"
        self.cursor.execute("truncate table TestNCLOBs")
        self.cursor.setinputsizes(val=oracledb.DB_TYPE_NVARCHAR)
        self.cursor.execute("""
                insert into TestNCLOBs (IntCol, NClobCol)
                values (1, :val)""",
                val=value)
        self.cursor.execute("select NCLOBCol from TestNCLOBs")
        nclob, = self.cursor.fetchone()
        self.cursor.setinputsizes(val=oracledb.DB_TYPE_NVARCHAR)
        self.cursor.execute("update TestNCLOBs set NCLOBCol = :val",
                            val=nclob.read() + value)
        self.cursor.execute("select NCLOBCol from TestNCLOBs")
        nclob, = self.cursor.fetchone()
        self.assertEqual(nclob.read(), value + value)

    def test_1916_nclob_indirect(self):
        "1916 - test binding and fetching NCLOB data (indirectly)"
        self.__perform_test("NCLOB", oracledb.DB_TYPE_LONG)

    def test_1917_nclob_operations(self):
        "1917 - test operations on NCLOBs"
        self.__test_lob_operations("NCLOB")

    def test_1918_temporary_lobs(self):
        "1918 - test temporary LOBs"
        self.cursor.execute("""
                select sys_context('USERENV', 'SID')
                from dual""")
        sid, = self.cursor.fetchone()
        temp_lobs = self.__get_temp_lobs(sid)
        with self.connection.cursor() as cursor:
            cursor.arraysize = 27
            self.assertEqual(temp_lobs, 0)
            cursor.execute("""
                    select extract(xmlcol, '/').getclobval()
                    from TestXML""")
            for lob, in cursor:
                value = lob.read()
                del lob
        temp_lobs = self.__get_temp_lobs(sid)
        self.assertEqual(temp_lobs, 0)

    def test_1919_assign_string_beyond_array_size(self):
        "1919 - test assign string to NCLOB beyond array size"
        nclobVar = self.cursor.var(oracledb.DB_TYPE_NCLOB)
        self.assertRaises(IndexError, nclobVar.setvalue, 1, "test char")

    def test_1920_supplemental_characters(self):
        "1920 - test read/write temporary LOBs using supplemental characters"
        self.cursor.execute("""
                select value
                from nls_database_parameters
                where parameter = 'NLS_CHARACTERSET'""")
        charset, = self.cursor.fetchone()
        if charset != "AL32UTF8":
            self.skipTest("Database character set must be AL32UTF8")
        supplemental_chars = "𠜎 𠜱 𠝹 𠱓 𠱸 𠲖 𠳏 𠳕 𠴕 𠵼 𠵿 𠸎 𠸏 𠹷 𠺝 " \
                "𠺢 𠻗 𠻹 𠻺 𠼭 𠼮 𠽌 𠾴 𠾼 𠿪 𡁜 𡁯 𡁵 𡁶 𡁻 𡃁 𡃉 𡇙 𢃇 " \
                "𢞵 𢫕 𢭃 𢯊 𢱑 𢱕 𢳂 𢴈 𢵌 𢵧 𢺳 𣲷 𤓓 𤶸 𤷪 𥄫 𦉘 𦟌 𦧲 " \
                "𦧺 𧨾 𨅝 𨈇 𨋢 𨳊 𨳍 𨳒 𩶘"
        self.cursor.execute("truncate table TestCLOBs")
        lob = self.connection.createlob(oracledb.DB_TYPE_CLOB)
        lob.write(supplemental_chars)
        self.cursor.execute("""
                insert into TestCLOBs
                (IntCol, ClobCol)
                values (1, :val)""",
                [lob])
        self.connection.commit()
        self.cursor.execute("select ClobCol from TestCLOBs")
        lob, = self.cursor.fetchone()
        self.assertEqual(lob.read(), supplemental_chars)

    def test_1921_plsql_auto_conversion_clob(self):
        "1921 - test automatic conversion to CLOB for PL/SQL"
        var = self.cursor.var(str, outconverter=lambda v: v[-15:])
        var.setvalue(0, "A" * 50000)
        self.cursor.execute("""
                declare
                    t_Clob          clob;
                begin
                    t_Clob := :data;
                    dbms_lob.copy(:data, t_Clob, 50000);
                    dbms_lob.writeappend(:data, 5, 'BBBBB');
                end;""",
                data=var)
        result = var.getvalue()
        self.assertEqual(result, "A" * 10 + "B" * 5)

    def test_1922_plsql_auto_conversion_nclob(self):
        "1922 - test automatic conversion to NCLOB for PL/SQL"
        var = self.cursor.var(oracledb.DB_TYPE_NCHAR,
                              outconverter=lambda v: v[-12:])
        var.setvalue(0, "N" * 51234)
        self.cursor.execute("""
                declare
                    t_Clob          nclob;
                begin
                    t_Clob := :data;
                    dbms_lob.copy(:data, t_Clob, 51234);
                    dbms_lob.writeappend(:data, 7, 'PPPPPPP');
                end;""",
                data=var)
        result = var.getvalue()
        self.assertEqual(result, "N" * 5 + "P" * 7)

    def test_1923_plsql_auto_conversion_blob(self):
        "1923 - test automatic conversion to BLOB for PL/SQL"
        var = self.cursor.var(bytes,
                              outconverter=lambda v: v[-14:])
        var.setvalue(0, b"L" * 52345)
        self.cursor.execute("""
                declare
                    t_Blob          blob;
                begin
                    t_Blob := :data;
                    dbms_lob.copy(:data, t_Blob, 52345);
                    dbms_lob.writeappend(:data, 6, '515151515151');
                end;""",
                data=var)
        result = var.getvalue()
        self.assertEqual(result, b"L" * 8 + b"Q" * 6)

    def test_1924_pickle_blob(self):
        "1924 - test pickling of BLOB"
        self.__test_pickle("BLOB")

    def test_1925_pickle_clob(self):
        "1925 - test pickling of CLOB"
        self.__test_pickle("CLOB")

    def test_1926_pickle_nclob(self):
        "1925 - test pickling of NCLOB"
        self.__test_pickle("NCLOB")

    def test_1927_fetch_blob_as_bytes(self):
        "1927 - test fetching BLOB as bytes"
        self.__test_fetch_lobs_direct("BLOB")

    def test_1928_fetch_clob_as_str(self):
        "1928 - test fetching CLOB as str"
        self.__test_fetch_lobs_direct("CLOB")

    def test_1929_fetch_nclob_as_str(self):
        "1929 - test fetching NCLOB as str"
        self.__test_fetch_lobs_direct("NCLOB")

    def test_1930_bind_order_blob(self):
        "1930 - test bind ordering with BLOB"
        self.__test_bind_ordering("BLOB")

    def test_1931_bind_order_clob(self):
        "1931 - test bind ordering with CLOB"
        self.__test_bind_ordering("CLOB")

    def test_1932_bind_order_nclob(self):
        "1932 - test bind ordering with NCLOB"
        self.__test_bind_ordering("NCLOB")

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2022, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# connect_params.pyx
#
# Cython file defining the base ConnectParams implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

# a set of alternative parameter names that are used in connect descriptors
# the key is the parameter name used in connect descriptors and the value is
# the key used in argument dictionaries (and stored in the parameter objects)
ALTERNATIVE_PARAM_NAMES = {
    "pool_connection_class": "cclass",
    "pool_purity": "purity",
    "server": "server_type",
    "transport_connect_timeout": "tcp_connect_timeout",
    "my_wallet_directory": "wallet_location"
}

# a set of parameter names in connect descriptors that are treated as
# containers (values are always lists)
CONTAINER_PARAM_NAMES = set((
    "address_list",
    "description",
    "address"
))

# regular expression used for determining if a connect string refers to an Easy
# Connect string or not
EASY_CONNECT_PATTERN = \
        "((?P<protocol>\w+)://)?(?P<host>[^:/]+)(:(?P<port>\d+)?)?/" \
        "(?P<service_name>[^:?]*)(:(?P<server_type>\w+))?"

# dictionary of tnsnames.ora files, indexed by the directory in which the file
# is found; the results are cached in order to avoid parsing a file multiple
# times; the modification time of the file is checked each time, though, to
# ensure that no changes were made since the last time that the file was read
# and parsed.
_tnsnames_files = {}

# internal default values
cdef str DEFAULT_PROTOCOL = "tcp"
cdef uint32_t DEFAULT_PORT = 1521
cdef double DEFAULT_TCP_CONNECT_TIMEOUT = 60


cdef int _add_container(dict args, str name, object value) except -1:
    """
    Adds a container to the arguments.
    """
    if name == "address" and "address_list" in args:
        value = dict(address=[value])
        name = "address_list"
    elif name == "address_list" and "address" in args:
        args[name] = [dict(address=[v]) for v in args["address"]]
        del args["address"]
    args.setdefault(name, []).append(value)


cdef dict _parse_connect_descriptor(str data, dict args):
    """
    Internal method which parses a connect descriptor containing name-value
    pairs in the form (KEY = VALUE), where the value can itself be another
    set of nested name-value pairs. A dictionary is returned containing
    these key value pairs.
    """
    if data[0] != "(" or data[-1] != ")":
        errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR, data=data)
    data = data[1:-1]
    pos = data.find("=")
    if pos < 0:
        errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR, data=data)
    name = data[:pos].strip().lower()
    data = data[pos + 1:].strip()
    if not data or not data.startswith("("):
        value = data
        if value and value[0] == '"' and value[-1] == '"':
            value = value[1:-1]
    else:
        value = {}
        while data:
            search_pos = 1
            num_opening_parens = 1
            num_closing_parens = 0
            while num_closing_parens < num_opening_parens:
                end_pos = data.find(")", search_pos)
                if end_pos < 0:
                    errors._raise_err(errors.ERR_INVALID_CONNECT_DESCRIPTOR,
                                      data=data)
                num_closing_parens += 1
                num_opening_parens += data.count("(", search_pos, end_pos)
                search_pos = end_pos + 1
            _parse_connect_descriptor(data[:end_pos + 1].strip(), value)
            data = data[end_pos + 1:].strip()
    name = ALTERNATIVE_PARAM_NAMES.get(name, name)
    if name in CONTAINER_PARAM_NAMES:
        _add_container(args, name, value)
    else:
        args[name] = value
    return args


cdef class ConnectParamsImpl:

    def __init__(self):
        cdef AddressList address_list
        self.stmtcachesize = defaults.stmtcachesize
        self.config_dir = defaults.config_dir
        self._default_description = Description()
        self._default_address = Address()
        self.description_list = DescriptionList()
        self.description_list.descriptions.append(self._default_description)
        self.debug_jdwp = os.getenv("ORA_DEBUG_JDWP")
        address_list = AddressList()
        address_list.addresses.append(self._default_address)
        self._default_description.address_lists.append(address_list)

    def set(self, dict args):
        """
        Sets the property values based on the supplied arguments. All values
        not supplied will be left unchanged.
        """
        self._external_handle = args.get("handle", 0)
        _set_str_param(args, "user", self)
        _set_str_param(args, "proxy_user", self)
        if self.proxy_user is None and self.user is not None:
            self.parse_user(self.user)
        self._set_password(args.get("password"))
        self._set_new_password(args.get("newpassword"))
        self._set_wallet_password(args.get("wallet_password"))
        _set_bool_param(args, "events", &self.events)
        _set_uint_param(args, "mode", &self.mode)
        _set_str_param(args, "edition", self)
        _set_str_param(args, "tag", self)
        _set_bool_param(args, "matchanytag", &self.matchanytag)
        _set_uint_param(args, "stmtcachesize", &self.stmtcachesize)
        _set_bool_param(args, "disable_oob", &self.disable_oob)
        _set_str_param(args, "debug_jdwp", self)
        _set_str_param(args, "config_dir", self)
        self.appcontext = args.get("appcontext")
        self.shardingkey = args.get("shardingkey")
        self.supershardingkey = args.get("supershardingkey")
        self._default_description.set_from_connect_data_args(args)
        self._default_description.set_from_description_args(args)
        self._default_description.set_from_security_args(args)
        self._default_address.set_from_args(args)
        _set_bool_param(args, "externalauth", &self.externalauth)
        self._set_access_token_param(args.get("access_token"))

    cdef int _check_credentials(self) except -1:
        """
        Check to see that credentials have been supplied: either a password or
        an access token.
        """
        if self._password is None and self._token is None \
                and self.access_token_callback is None:
            errors._raise_err(errors.ERR_NO_CREDENTIALS)

    cdef int _copy(self, ConnectParamsImpl other_params) except -1:
        """
        Internal method for copying attributes from another set of parameters.
        """
        self.config_dir = other_params.config_dir
        self.user = other_params.user
        self.proxy_user = other_params.proxy_user
        self.events = other_params.events
        self.externalauth = other_params.externalauth
        self.mode = other_params.mode
        self.edition = other_params.edition
        self.appcontext = other_params.appcontext
        self.tag = other_params.tag
        self.matchanytag = other_params.matchanytag
        self.shardingkey = other_params.shardingkey
        self.supershardingkey = other_params.supershardingkey
        self.stmtcachesize = other_params.stmtcachesize
        self.disable_oob = other_params.disable_oob
        self.debug_jdwp = other_params.debug_jdwp
        self.description_list = other_params.description_list
        self.access_token_callback = other_params.access_token_callback
        self.access_token_expires = other_params.access_token_expires
        self._external_handle = other_params._external_handle
        self._default_description = other_params._default_description
        self._default_address = other_params._default_address
        self._password = other_params._password
        self._password_obfuscator = other_params._password_obfuscator
        self._new_password = other_params._new_password
        self._new_password_obfuscator = other_params._new_password_obfuscator
        self._wallet_password = other_params._wallet_password
        self._wallet_password_obfuscator = \
                other_params._wallet_password_obfuscator
        self._token = other_params._token
        self._token_obfuscator = other_params._token_obfuscator
        self._private_key = other_params._private_key
        self._private_key_obfuscator = other_params._private_key_obfuscator

    cdef bytes _get_new_password(self):
        """
        Returns the new password, after removing the obfuscation.
        """
        if self._new_password is not None:
            return bytes(self._xor_bytes(self._new_password,
                                         self._new_password_obfuscator))

    cdef bytearray _get_obfuscator(self, str secret_value):
        """
        Return a byte array suitable for obfuscating the specified secret
        value.
        """
        return bytearray(secrets.token_bytes(len(secret_value.encode())))

    cdef bytes _get_password(self):
        """
        Returns the password, after removing the obfuscation.
        """
        if self._password is not None:
            return bytes(self._xor_bytes(self._password,
                                         self._password_obfuscator))

    cdef str _get_private_key(self):
        """
        Returns the private key, after removing the obfuscation.
        """
        if self._private_key is not None:
            return self._xor_bytes(self._private_key,
                                   self._private_key_obfuscator).decode()

    cdef TnsnamesFile _get_tnsnames_file(self):
        """
        Return a tnsnames file, if one is present, or None if one is not. If
        the file was previously loaded, the modification time is checked and,
        if unchanged, the previous file is returned.
        """
        cdef TnsnamesFile tnsnames_file
        if self.config_dir is None:
            errors._raise_err(errors.ERR_NO_CONFIG_DIR)
        file_name = os.path.join(self.config_dir, "tnsnames.ora")
        tnsnames_file = _tnsnames_files.get(self.config_dir)
        try:
            stat_info = os.stat(file_name)
        except:
            if tnsnames_file is not None:
                del _tnsnames_files[self.config_dir]
            errors._raise_err(errors.ERR_TNS_NAMES_FILE_MISSING,
                              config_dir=self.config_dir)
        if tnsnames_file is not None \
                and tnsnames_file.mtime == stat_info.st_mtime:
            return tnsnames_file
        tnsnames_file = TnsnamesFile(file_name, stat_info.st_mtime)
        tnsnames_file.read()
        _tnsnames_files[self.config_dir] = tnsnames_file
        return tnsnames_file

    cdef str _get_token(self):
        """
        Returns the token, after removing the obfuscation.

        If a callable has been registered and there is no token stored yet, the
        callable will be invoked with the refresh parameter set to False. If
        the token returned by the callable is expired, the callable will be
        invoked a second time with the refresh parameter set to True. If this
        token is also expired, an exception will be raised.

        If the stored token has expired and no callable has been registered, an
        exception will be raised; otherwise, the callable will be invoked with
        the refresh parameter set to True.
        """
        cdef:
            object returned_val, current_date = datetime.datetime.utcnow()
            bint expired
        if self._token is None and self.access_token_callback is not None:
            returned_val = self.access_token_callback(False)
            self._set_access_token(returned_val,
                                   errors.ERR_INVALID_ACCESS_TOKEN_RETURNED)
        expired = self.access_token_expires < current_date
        if expired and self.access_token_callback is not None:
            returned_val = self.access_token_callback(True)
            self._set_access_token(returned_val,
                                   errors.ERR_INVALID_ACCESS_TOKEN_RETURNED)
            expired = self.access_token_expires < current_date
        if expired:
            errors._raise_err(errors.ERR_EXPIRED_ACCESS_TOKEN)
        return self._xor_bytes(self._token, self._token_obfuscator).decode()

    cdef object _get_token_expires(self, str token):
        """
        Gets the expiry date from the token.
        """
        cdef:
            str header_seg
            dict header
            int num_pad
        header_seg = token.split(".")[1]
        num_pad = len(header_seg) % 4
        if num_pad != 0:
            header_seg += '=' * num_pad
        header = json.loads(base64.b64decode(header_seg))
        return datetime.datetime.utcfromtimestamp(header["exp"])

    cdef str _get_wallet_password(self):
        """
        Returns the wallet password, after removing the obfuscation.
        """
        if self._wallet_password is not None:
            return self._xor_bytes(self._wallet_password,
                                   self._wallet_password_obfuscator).decode()

    cdef int _parse_connect_string(self, str connect_string) except -1:
        """
        Internal method for parsing a connect string.
        """
        cdef:
            TnsnamesFile tnsnames_file
            Description description
            Address address
            dict args = {}
            str name

        # if a connect string starts with an opening parenthesis it is assumed
        # to be a full connect descriptor
        if connect_string.startswith("("):
            _parse_connect_descriptor(connect_string, args)
            return self._process_connect_descriptor(args)

        # otherwise, see if the connect string is an EasyConnect string
        m = re.search(EASY_CONNECT_PATTERN, connect_string)
        if m is not None:

            # build up arguments
            args = m.groupdict()
            connect_string = connect_string[m.end():]
            params_pos = connect_string.find("?")
            if params_pos >= 0:
                params = connect_string[params_pos + 1:]
                for part in params.split("&"):
                    name, value = [s.strip() for s in part.split("=", 1)]
                    name = name.lower()
                    name = ALTERNATIVE_PARAM_NAMES.get(name, name)
                    if value.startswith('"') and value.endswith('"'):
                        value = value[1:-1]
                    args[name] = value

            # create description list
            address = self._default_address.copy()
            address.set_from_args(args)
            description = self._default_description.copy()
            description.set_from_connect_data_args(args)
            description.set_from_description_args(args)
            description.set_from_security_args(args)
            description.address_lists = [AddressList()]
            description.address_lists[0].addresses.append(address)
            self.description_list = DescriptionList()
            self.description_list.descriptions.append(description)

        # otherwise, see if the name is a connect alias in a tnsnames.ora
        # configuration file
        else:
            tnsnames_file = self._get_tnsnames_file()
            name = connect_string
            connect_string = tnsnames_file.entries.get(name.upper())
            if connect_string is None:
                errors._raise_err(errors.ERR_TNS_ENTRY_NOT_FOUND, name=name,
                                  file_name=tnsnames_file.file_name)
            _parse_connect_descriptor(connect_string, args)
            self._process_connect_descriptor(args)

    cdef int _process_connect_descriptor(self, dict args) except -1:
        """
        Internal method used for processing the parsed connect descriptor into
        the set of DescriptionList, Description, AddressList and Address
        container objects.
        """
        cdef:
            AddressList address_list
            Description description
            Address address
        self.description_list = DescriptionList()
        list_args = args.get("description_list")
        if list_args is not None:
            self.description_list.set_from_args(list_args)
        else:
            list_args = args
        for desc_args in list_args.get("description", [list_args]):
            description = self._default_description.copy()
            description.set_from_description_args(desc_args)
            self.description_list.descriptions.append(description)
            sub_args = desc_args.get("connect_data")
            if sub_args is not None:
                description.set_from_connect_data_args(sub_args)
            sub_args = desc_args.get("security")
            if sub_args is not None:
                description.set_from_security_args(sub_args)
            for list_args in desc_args.get("address_list", [desc_args]):
                address_list = AddressList()
                address_list.set_from_args(list_args)
                description.address_lists.append(address_list)
                for addr_args in list_args.get("address", []):
                    address = self._default_address.copy()
                    address.set_from_args(addr_args)
                    address_list.addresses.append(address)

    cdef int _set_access_token(self, object val, int error_num) except -1:
        """
        Sets the access token either supplied directly by the user or
        indirectly via a callback.
        """
        cdef:
            str token, private_key = None
            object token_expires
        if isinstance(val, tuple) and len(val) == 2:
            token, private_key = val
            if token is None or private_key is None:
                errors._raise_err(error_num)
        elif isinstance(val, str):
            token = val
        else:
            errors._raise_err(error_num)
        try:
            token_expires = self._get_token_expires(token)
        except Exception as e:
            errors._raise_err(error_num, cause=e)
        self._token_obfuscator = self._get_obfuscator(token)
        self._token = self._xor_bytes(bytearray(token.encode()),
                                                self._token_obfuscator)
        if private_key is not None:
            self._private_key_obfuscator = self._get_obfuscator(private_key)
            self._private_key = self._xor_bytes(bytearray(private_key.encode()),
                                                self._private_key_obfuscator)
        self.access_token_expires = token_expires

    cdef int _set_access_token_param(self, object val) except -1:
        """
        Sets the access token parameter.
        """
        if val is not None:
            if callable(val):
                self.access_token_callback = val
            else:
                self._set_access_token(val,
                                       errors.ERR_INVALID_ACCESS_TOKEN_PARAM)

    cdef int _set_new_password(self, str password) except -1:
        """
        Sets the new password on the instance after first obfuscating it.
        """
        if password is not None:
            self._new_password_obfuscator = self._get_obfuscator(password)
            self._new_password = self._xor_bytes(bytearray(password.encode()),
                                                 self._new_password_obfuscator)

    cdef int _set_password(self, str password) except -1:
        """
        Sets the password on the instance after first obfuscating it.
        """
        if password is not None:
            self._password_obfuscator = self._get_obfuscator(password)
            self._password = self._xor_bytes(bytearray(password.encode()),
                                             self._password_obfuscator)

    cdef int _set_wallet_password(self, str password) except -1:
        """
        Sets the wallet password on the instance after first obfuscating it.
        """
        if password is not None:
            self._wallet_password_obfuscator = self._get_obfuscator(password)
            self._wallet_password = \
                    self._xor_bytes(bytearray(password.encode()),
                                    self._wallet_password_obfuscator)

    cdef bytearray _xor_bytes(self, bytearray a, bytearray b):
        """
        Perform an XOR of two byte arrays as a means of obfuscating a password
        that is stored on the class. It is assumed that the byte arrays are of
        the same length.
        """
        cdef:
            ssize_t length, i
            bytearray result
        length = len(a)
        result = bytearray(length)
        for i in range(length):
            result[i] = a[i] ^ b[i]
        return result

    def copy(self):
        """
        Creates a copy of the connection parameters and returns it.
        """
        cdef ConnectParamsImpl new_params
        new_params = ConnectParamsImpl.__new__(ConnectParamsImpl)
        new_params._copy(self)
        return new_params

    def _get_addresses(self):
        """
        Return a list of the stored addresses.
        """
        cdef:
            AddressList addr_list
            Description desc
            Address addr
        return [addr for desc in self.description_list.descriptions \
                for addr_list in desc.address_lists \
                for addr in addr_list.addresses]

    def get_connect_string(self):
        """
        Returns a connect string generated from the parameters.
        """
        return self.description_list.build_connect_string()

    def get_full_user(self):
        """
        Internal method used for getting the full user (including any proxy
        user) which is used in thick mode exlusively and which is used in
        the repr() methods for Connection and ConnectionPool.
        """
        if self.proxy_user is not None:
            return f"{self.user}[{self.proxy_user}]"
        return self.user

    def parse_connect_string(self, str connect_string):
        """
        Internal method for parsing the connect string.
        """
        connect_string = connect_string.strip()
        try:
            self._parse_connect_string(connect_string)
        except exceptions.Error:
            raise
        except Exception as e:
            errors._raise_err(errors.ERR_CANNOT_PARSE_CONNECT_STRING, cause=e,
                              data=connect_string)

    def parse_dsn_with_credentials(self, str dsn):
        """
        Parse a dsn (data source name) string supplied by the user. This can be
        in the form user/password@connect_string or it can be in the form
        user/password. The user, password and connect string are returned in a
        3-tuple.
        """
        pos = dsn.rfind("@")
        if pos >= 0:
            credentials = dsn[:pos]
            connect_string = dsn[pos + 1:] or None
        else:
            credentials = dsn
            connect_string = None
        pos = credentials.find("/")
        if pos >= 0:
            user = credentials[:pos] or None
            password = credentials[pos + 1:] or None
        elif connect_string is None:
            connect_string = dsn or None
            user = password = None
        else:
            user = credentials or None
            password = None
        return (user, password, connect_string)

    def parse_user(self, str user):
        """
        Parses a user string into its component parts, if applicable. The user
        string may be in the form user[proxy_user] or it may simply be a simple
        user string.
        """
        start_pos = user.find("[")
        if start_pos > 0 and user.endswith("]"):
            self.proxy_user = user[start_pos + 1:-1]
            self.user = user[:start_pos]
        else:
            self.user = user

    def process_args(self, str dsn, dict kwargs, bint thin):
        """
        Processes the arguments to connect() and create_pool().

            - the keyword arguments are set
            - if no user was specified in the keyword arguments and a dsn is
              specified, it is parsed to determine the user, password and
              connect string and the user and password are stored
            - in thin mode, the connect string is then parsed into its
              components and stored
            - if no dsn was specified, one is built from the components
            - the connect string is returned
        """
        if kwargs:
            self.set(kwargs)
        if self.user is None and not self.externalauth and dsn is not None:
            user, password, dsn = self.parse_dsn_with_credentials(dsn)
            self.set(dict(user=user, password=password))
        if dsn is not None and thin:
            self.parse_connect_string(dsn)
        if dsn is None:
            dsn = self.get_connect_string()
        return dsn


cdef class Address:
    """
    Internal class used to hold parameters for an address used to create a
    connection to the database.
    """

    def __init__(self):
        self.protocol = DEFAULT_PROTOCOL
        self.port = DEFAULT_PORT

    cdef str build_connect_string(self):
        """
        Build a connect string from the components. If no host is specified,
        None is returned (used for bequeath connections).
        """
        if self.host is not None:
            parts = [f"(PROTOCOL={self.protocol})",
                     f"(HOST={self.host})",
                     f"(PORT={self.port})"]
            if self.https_proxy is not None:
                parts.append(f"(HTTPS_PROXY={self.https_proxy})")
            if self.https_proxy_port != 0:
                parts.append(f"(HTTPS_PROXY_PORT={self.https_proxy_port})")
            return f'(ADDRESS={"".join(parts)})'

    def copy(self):
        """
        Creates a copy of the address and returns it.
        """
        cdef Address address = Address.__new__(Address)
        address.host = self.host
        address.port = self.port
        address.protocol = self.protocol
        address.https_proxy = self.https_proxy
        address.https_proxy_port = self.https_proxy_port
        return address

    @classmethod
    def from_args(cls, dict args):
        """
        Creates an address and sets the arguments before returning it. This is
        used within connect descriptors containing address lists.
        """
        address = cls()
        address.set_from_args(args)
        return address

    def set_from_args(self, dict args):
        """
        Sets parameter values from an argument dictionary or an (ADDRESS) node
        in a connect descriptor.
        """
        _set_str_param(args, "host", self)
        _set_uint_param(args, "port", &self.port)
        _set_protocol_param(args, "protocol", self)
        _set_str_param(args, "https_proxy", self)
        _set_uint_param(args, "https_proxy_port", &self.https_proxy_port)


cdef class AddressList:
    """
    Internal class used to hold address list parameters and a list of addresses
    used to create connections to the database.
    """

    def __init__(self):
        self.addresses = []

    cdef bint _uses_tcps(self):
        """
        Returns a boolean indicating if any of the addresses in the address
        list use the protocol TCPS.
        """
        cdef Address address
        for address in self.addresses:
            if address.protocol == "tcps":
                return True
        return False

    cdef str build_connect_string(self):
        """
        Build a connect string from the components.
        """
        cdef Address a
        parts = [a.build_connect_string() for a in self.addresses]
        if len(parts) == 1:
            return parts[0]
        return f'(ADDRESS_LIST={"".join(parts)})'

    def set_from_args(self, dict args):
        """
        Set paramter values from an argument dictionary or an (ADDRESS_LIST)
        node in a connect descriptor.
        """
        _set_bool_param(args, "load_balance", &self.load_balance)
        _set_bool_param(args, "source_route", &self.source_route)


cdef class Description:
    """
    Internal class used to hold description parameters.
    """

    def __init__(self):
        self.address_lists = []
        self.tcp_connect_timeout = DEFAULT_TCP_CONNECT_TIMEOUT
        self.ssl_server_dn_match = True

    cdef str _build_duration_str(self, double value):
        """
        Build up the value to display for a duration in the connect string.
        This must be an integer with the units following it.
        """
        cdef int value_int, value_minutes
        value_int = <int> value
        if value != value_int:
            return f"{int(value * 1000)}ms"
        value_minutes = (value_int // 60)
        if value_minutes * 60 == value_int:
            return f"{value_minutes}min"
        return f"{value_int}"

    cdef str build_connect_string(self, str cid=None):
        """
        Build a connect string from the components.
        """
        cdef:
            AddressList address_list
            list parts, temp_parts
            bint uses_tcps = False
            str temp

        # build top-level description parts
        parts = []
        if self.load_balance:
            parts.append("(LOAD_BALANCE=ON)")
        if self.source_route:
            parts.append("(SOURCE_ROUTE=ON)")
        if self.retry_count != 0:
            parts.append(f"(RETRY_COUNT={self.retry_count})")
        if self.retry_delay != 0:
            parts.append(f"(RETRY_DELAY={self.retry_delay})")
        if self.expire_time != 0:
            parts.append(f"(EXPIRE_TIME={self.expire_time})")
        if self.tcp_connect_timeout != DEFAULT_TCP_CONNECT_TIMEOUT:
            temp = self._build_duration_str(self.tcp_connect_timeout)
            parts.append(f"(TRANSPORT_CONNECT_TIMEOUT={temp})")

        # add address lists, but if the address list contains only a single
        # entry and that entry does not have a host, the other parts aren't
        # relevant anyway!
        for address_list in self.address_lists:
            temp = address_list.build_connect_string()
            if temp is None:
                return None
            parts.append(temp)
            if not uses_tcps:
                uses_tcps = address_list._uses_tcps()

        # build connect data segment
        temp_parts = []
        if self.service_name is not None:
            temp_parts.append(f"(SERVICE_NAME={self.service_name})")
        elif self.sid is not None:
            temp_parts.append(f"(SID={self.sid})")
        if self.server_type is not None:
            temp_parts.append(f"(SERVER={self.server_type})")
        if self.cclass is not None:
            temp_parts.append(f"(POOL_CONNECTION_CLASS={self.cclass})")
        if self.purity != 0:
            temp_parts.append(f"(POOL_PURITY={self.purity})")
        if cid is not None:
            temp_parts.append(f"(CID={cid})")
        if temp_parts:
            parts.append(f'(CONNECT_DATA={"".join(temp_parts)})')

        # build security segment, if applicable
        if uses_tcps:
            temp_parts = []
            if self.ssl_server_dn_match:
                temp_parts.append("(SSL_SERVER_DN_MATCH=ON)")
            if self.ssl_server_cert_dn is not None:
                temp = f"(SSL_SERVER_CERT_DN={self.ssl_server_cert_dn})"
                temp_parts.append(temp)
            if self.wallet_location is not None:
                temp = f"(MY_WALLET_DIRECTORY={self.wallet_location})"
                temp_parts.append(temp)
            parts.append(f'(SECURITY={"".join(temp_parts)})')

        return f'(DESCRIPTION={"".join(parts)})'

    def copy(self):
        """
        Creates a copy of the description (except for the address lists) and
        returns it.
        """
        cdef Description description = Description.__new__(Description)
        description.address_lists = []
        description.service_name = self.service_name
        description.sid = self.sid
        description.server_type = self.server_type
        description.cclass = self.cclass
        description.purity = self.purity
        description.expire_time = self.expire_time
        description.load_balance = self.load_balance
        description.source_route = self.source_route
        description.retry_count = self.retry_count
        description.retry_delay = self.retry_delay
        description.tcp_connect_timeout = self.tcp_connect_timeout
        description.ssl_server_dn_match = self.ssl_server_dn_match
        description.ssl_server_cert_dn = self.ssl_server_cert_dn
        description.wallet_location = self.wallet_location
        return description

    def set_from_connect_data_args(self, dict args):
        """
        Set parameter values from an argument dictionary or a (CONNECT_DATA)
        node in a connect descriptor.
        """
        _set_str_param(args, "service_name", self)
        _set_str_param(args, "sid", self)
        _set_server_type_param(args, "server_type", self)
        _set_str_param(args, "cclass", self)
        _set_purity_param(args, "purity", &self.purity)

    def set_from_description_args(self, dict args):
        """
        Set parameter values from an argument dictionary or a (DESCRIPTION)
        node in a connect descriptor.
        """
        cdef Address address
        _set_uint_param(args, "expire_time", &self.expire_time)
        _set_bool_param(args, "load_balance", &self.load_balance)
        _set_bool_param(args, "source_route", &self.source_route)
        _set_uint_param(args, "retry_count", &self.retry_count)
        _set_uint_param(args, "retry_delay", &self.retry_delay)
        _set_duration_param(args, "tcp_connect_timeout",
                            &self.tcp_connect_timeout)

    def set_from_security_args(self, dict args):
        """
        Set parameter values from an argument dictionary or a (SECURITY)
        node in a connect descriptor.
        """
        _set_bool_param(args, "ssl_server_dn_match", &self.ssl_server_dn_match)
        _set_str_param(args, "ssl_server_cert_dn", self)
        _set_str_param(args, "wallet_location", self)


cdef class DescriptionList:
    """
    Internal class used to hold description list parameters and a list of
    descriptions.
    """

    def __init__(self):
        self.descriptions = []

    cdef str build_connect_string(self):
        """
        Build a connect string from the components.
        """
        cdef:
            Description d
            list parts
        parts = [d.build_connect_string() for d in self.descriptions]
        if len(parts) == 1:
            return parts[0]
        return f'(DESCIPTION_LIST={"".join(parts)})'

    def set_from_args(self, dict args):
        """
        Set paramter values from an argument dictionary or a (DESCRIPTION_LIST)
        node in a connect descriptor.
        """
        _set_bool_param(args, "load_balance", &self.load_balance)
        _set_bool_param(args, "source_route", &self.source_route)


cdef class TnsnamesFile:
    """
    Internal class used to parse and retain connect descriptor entries found in
    a tnsnames.ora file.
    """

    def __init__(self, str file_name, int mtime):
        self.file_name = file_name
        self.mtime = mtime
        self.entries = {}

    def read(self):
        """
        Read and parse the file and retain the connect descriptors found inside
        the file.
        """
        with open(self.file_name) as f:
            entry_names = None
            for line in f:
                line = line.strip()
                pos = line.find("#")
                if pos >= 0:
                    line = line[:pos]
                if not line:
                    continue
                if entry_names is None:
                    pos = line.find("=")
                    if pos < 0:
                        continue
                    entry_names = [s.strip() for s in line[:pos].split(",")]
                    entry_lines = []
                    num_parens = 0
                    line = line[pos+1:].strip()
                if line:
                    num_parens += line.count("(") - line.count(")")
                    entry_lines.append(line)
                if entry_lines and num_parens <= 0:
                    descriptor = "".join(entry_lines)
                    for name in entry_names:
                        self.entries[name.upper()] = descriptor
                    entry_names = None



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
4100 - Module for testing the methods for calling stored procedures and
functions (callproc() and callfunc())
"""

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def test_4100_callproc(self):
        "4100 - test executing a stored procedure"
        var = self.cursor.var(oracledb.NUMBER)
        results = self.cursor.callproc("proc_Test", ("hi", 5, var))
        self.assertEqual(results, ["hi", 10, 2.0])

    def test_4101_callproc_all_keywords(self):
        "4101 - test executing a stored procedure with all args keyword args"
        inout_value = self.cursor.var(oracledb.NUMBER)
        inout_value.setvalue(0, 5)
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(a_InOutValue=inout_value, a_InValue="hi",
                      a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test", [], kwargs)
        self.assertEqual(results, [])
        self.assertEqual(inout_value.getvalue(), 10)
        self.assertEqual(out_value.getvalue(), 2.0)

    def test_4102_callproc_only_last_keyword(self):
        "4102 - test executing a stored procedure with last arg as keyword arg"
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test", ("hi", 5), kwargs)
        self.assertEqual(results, ["hi", 10])
        self.assertEqual(out_value.getvalue(), 2.0)

    def test_4103_callproc_repeated_keyword_parameters(self):
        "4103 - test executing a stored procedure, repeated keyword arg"
        kwargs = dict(a_InValue="hi",
                      a_OutValue=self.cursor.var(oracledb.NUMBER))
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-06550:",
                               self.cursor.callproc, "proc_Test", ("hi", 5),
                               kwargs)

    def test_4104_callproc_no_args(self):
        "4104 - test executing a stored procedure without any arguments"
        results = self.cursor.callproc("proc_TestNoArgs")
        self.assertEqual(results, [])

    def test_4105_callfunc(self):
        "4105 - test executing a stored function"
        results = self.cursor.callfunc("func_Test", oracledb.NUMBER, ("hi", 5))
        self.assertEqual(results, 7)

    def test_4106_callfunc_no_args(self):
        "4106 - test executing a stored function without any arguments"
        results = self.cursor.callfunc("func_TestNoArgs", oracledb.NUMBER)
        self.assertEqual(results, 712)

    def test_4107_callfunc_negative(self):
        "4107 - test executing a stored function with wrong parameters"
        func_name = "func_Test"
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2007:",
                               self.cursor.callfunc, oracledb.NUMBER,
                               func_name, ("hi", 5))
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-06550:",
                               self.cursor.callfunc, func_name,
                               oracledb.NUMBER, ("hi", 5, 7))
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2012:",
                               self.cursor.callfunc, func_name,
                               oracledb.NUMBER, "hi", 7)
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-06502:",
                               self.cursor.callfunc, func_name,
                               oracledb.NUMBER, [5, "hi"])
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-06550:",
                               self.cursor.callfunc, func_name,
                               oracledb.NUMBER)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2012:",
                               self.cursor.callfunc, func_name,
                               oracledb.NUMBER, 5)

    def test_4108_keywordParameters_deprecation(self):
        "4108 - test to verify keywordParameters is deprecated"
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(a_OutValue=out_value)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               self.cursor.callproc, "proc_Test", ("hi", 5),
                               kwargs, keywordParameters=kwargs)
        extra_amount = self.cursor.var(oracledb.NUMBER)
        extra_amount.setvalue(0, 5)
        kwargs = dict(a_ExtraAmount=extra_amount, a_String="hi")
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               self.cursor.callfunc, "func_Test",
                               oracledb.NUMBER, [], kwargs,
                               keywordParameters=kwargs)

    def test_4109_keyword_args_with_invalid_type(self):
        "4109 - test error for keyword args with invalid type"
        kwargs = [5]
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2013:",
                               self.cursor.callproc, "proc_Test", [], kwargs)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2013:",
                               self.cursor.callfunc, "func_Test",
                               oracledb.NUMBER, [], kwargs)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2022, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# dbobject.pyx
#
# Cython file defining the thin DbObjectType, DbObjectAttr and DbObject
# implementation classes (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class DbObjectPickleBuffer(GrowableBuffer):

    cdef int _read_raw_bytes_and_length(self, const char_type **ptr,
                                        ssize_t *num_bytes) except -1:
        """
        Helper function that processes the length (if needed) and then acquires
        the specified number of bytes from the buffer.
        """
        cdef uint32_t extended_num_bytes
        if num_bytes[0] == TNS_LONG_LENGTH_INDICATOR:
            self.read_uint32(&extended_num_bytes)
            num_bytes[0] = <ssize_t> extended_num_bytes
        ptr[0] = self._get_raw(num_bytes[0])

    cdef int _write_raw_bytes_and_length(self, const char_type *ptr,
                                         ssize_t num_bytes) except -1:
        """
        Helper function that writes the length in the format required before
        writing the bytes.
        """
        self.write_length(num_bytes)
        self.write_raw(ptr, <uint32_t> num_bytes)

    cdef int get_is_atomic_null(self, bint* is_null) except -1:
        """
        Reads the next byte and checks to see if the value is atomically null.
        If not, the byte is returned to the buffer for further processing.
        """
        cdef uint8_t value
        self.read_ub1(&value)
        if value in (TNS_OBJ_ATOMIC_NULL, TNS_NULL_LENGTH_INDICATOR):
            is_null[0] = True
        else:
            is_null[0] = False
            self._pos -= 1

    cdef int read_header(self, uint8_t* flags, uint8_t *version) except -1:
        """
        Reads the header of the pickled data.
        """
        cdef:
            uint32_t prefix_seg_length
            uint8_t tmp
        self.read_ub1(flags)
        self.read_ub1(version)
        self.skip_length()
        if flags[0] & TNS_OBJ_NO_PREFIX_SEG:
            return 0
        self.read_length(&prefix_seg_length)
        self.skip_raw_bytes(prefix_seg_length)

    cdef int read_length(self, uint32_t *length) except -1:
        """
        Read the length from the buffer. This will be a single byte, unless the
        value meets or exceeds TNS_LONG_LENGTH_INDICATOR. In that case, the
        value is stored as a 4-byte integer.
        """
        cdef uint8_t short_length
        self.read_ub1(&short_length)
        if short_length == TNS_LONG_LENGTH_INDICATOR:
            self.read_uint32(length)
        else:
            length[0] = short_length

    cdef int skip_length(self) except -1:
        """
        Skips the length instead of reading it from the buffer.
        """
        cdef uint8_t short_length
        self.read_ub1(&short_length)
        if short_length == TNS_LONG_LENGTH_INDICATOR:
            self.skip_raw_bytes(4)

    cdef int write_header(self, ThinDbObjectImpl obj_impl) except -1:
        """
        Writes the header of the pickled data. Since the size is unknown at
        this point, zero is written initially and the actual size is written
        later.
        """
        cdef ThinDbObjectTypeImpl typ_impl = obj_impl.type
        self.write_uint8(obj_impl.image_flags)
        self.write_uint8(obj_impl.image_version)
        self.write_uint8(TNS_LONG_LENGTH_INDICATOR)
        self.write_uint32(0)
        if typ_impl.is_collection:
            self.write_uint8(1)             # length of prefix segment
            self.write_uint8(1)             # prefix segment contents

    cdef int write_length(self, ssize_t length) except -1:
        """
        Writes the length to the buffer.
        """
        if length <= TNS_OBJ_MAX_SHORT_LENGTH:
            self.write_uint8(<uint8_t> length)
        else:
            self.write_uint8(TNS_LONG_LENGTH_INDICATOR)
            self.write_uint32(<uint32_t> length)


@cython.final
cdef class TDSBuffer(Buffer):
    pass


cdef class ThinDbObjectImpl(BaseDbObjectImpl):
    cdef:
        uint8_t image_flags, image_version
        bytes toid, oid, packed_data
        uint32_t num_elements
        dict unpacked_assoc_array
        list unpacked_assoc_keys
        dict unpacked_attrs
        list unpacked_array
        uint16_t flags

    cdef inline int _ensure_assoc_keys(self) except -1:
        """
        Ensure that the keys for the associative array have been calculated.
        PL/SQL associative arrays keep their keys in sorted order so this must
        be calculated when indices are required.
        """
        if self.unpacked_assoc_keys is None:
            self.unpacked_assoc_keys = list(sorted(self.unpacked_assoc_array))

    cdef inline int _ensure_unpacked(self) except -1:
        """
        Ensure that the data has been unpacked.
        """
        if self.packed_data is not None:
            self._unpack_data()

    cdef bytes _get_packed_data(self):
        """
        Returns the packed data for the object. This will either be the value
        retrieved from the database or generated packed data (for new objects
        and those that have had their data unpacked already).
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl = self.type
            DbObjectPickleBuffer buf
            ssize_t size
        if self.packed_data is not None:
            return self.packed_data
        buf = DbObjectPickleBuffer.__new__(DbObjectPickleBuffer)
        buf._initialize(TNS_CHUNK_SIZE)
        buf.write_header(self)
        self._pack_data(buf)
        size = buf._pos
        buf.skip_to(3)
        buf.write_uint32(size)
        return buf._data[:size]

    cdef int _pack_data(self, DbObjectPickleBuffer buf) except -1:
        """
        Packs the data from the object into the buffer.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl = self.type
            ThinDbObjectAttrImpl attr
            int32_t index
            object value
        if typ_impl.is_collection:
            buf.write_uint8(typ_impl.collection_flags)
            if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                self._ensure_assoc_keys()
                buf.write_length(len(self.unpacked_assoc_keys))
                for index in self.unpacked_assoc_keys:
                    buf.write_uint32(<uint32_t> index)
                    self._pack_value(buf, typ_impl.element_dbtype,
                                     typ_impl.element_objtype,
                                     self.unpacked_assoc_array[index])
            else:
                buf.write_length(len(self.unpacked_array))
                for value in self.unpacked_array:
                    self._pack_value(buf, typ_impl.element_dbtype,
                                    typ_impl.element_objtype, value)
        else:
            for attr in typ_impl.attrs:
                self._pack_value(buf, attr.dbtype, attr.objtype,
                                 self.unpacked_attrs[attr.name])

    cdef int _pack_value(self, DbObjectPickleBuffer buf,
                         DbType dbtype, ThinDbObjectTypeImpl objtype,
                         object value) except -1:
        """
        Packs a value into the buffer. At this point it is assumed that the
        value matches the correct type.
        """
        cdef:
            uint8_t ora_type_num = dbtype._ora_type_num
            ThinDbObjectImpl obj_impl
            bytes temp_bytes
        if value is None:
            if objtype is not None and not objtype.is_collection:
                buf.write_uint8(TNS_OBJ_ATOMIC_NULL)
            else:
                buf.write_uint8(TNS_NULL_LENGTH_INDICATOR)
        elif ora_type_num in (TNS_DATA_TYPE_CHAR, TNS_DATA_TYPE_VARCHAR):
            if dbtype._csfrm == TNS_CS_IMPLICIT:
                temp_bytes = (<str> value).encode()
            else:
                temp_bytes = (<str> value).encode(TNS_ENCODING_UTF16)
            buf.write_bytes_with_length(temp_bytes)
        elif ora_type_num == TNS_DATA_TYPE_NUMBER:
            temp_bytes = (<str> cpython.PyObject_Str(value)).encode()
            buf.write_oracle_number(temp_bytes)
        elif ora_type_num == TNS_DATA_TYPE_BINARY_INTEGER:
            buf.write_uint8(4)
            buf.write_uint32(<uint32_t> value)
        elif ora_type_num == TNS_DATA_TYPE_RAW:
            buf.write_bytes_with_length(value)
        elif ora_type_num == TNS_DATA_TYPE_BINARY_DOUBLE:
            buf.write_binary_double(value)
        elif ora_type_num == TNS_DATA_TYPE_BINARY_FLOAT:
            buf.write_binary_float(value)
        elif ora_type_num == TNS_DATA_TYPE_BOOLEAN:
            buf.write_uint8(4)
            buf.write_uint32(value)
        elif ora_type_num in (TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_TIMESTAMP,
                              TNS_DATA_TYPE_TIMESTAMP_TZ,
                              TNS_DATA_TYPE_TIMESTAMP_LTZ):
            buf.write_oracle_date(value, dbtype._buffer_size_factor)
        elif ora_type_num in (TNS_DATA_TYPE_CLOB, TNS_DATA_TYPE_BLOB):
            buf.write_lob(value._impl)
        elif ora_type_num == TNS_DATA_TYPE_INT_NAMED:
            obj_impl = value._impl
            if self.type.is_collection or obj_impl.type.is_collection:
                temp_bytes = obj_impl._get_packed_data()
                buf.write_bytes_with_length(temp_bytes)
            else:
                obj_impl._pack_data(buf)
        else:
            errors._raise_err(errors.ERR_DB_TYPE_NOT_SUPPORTED,
                              name=dbtype.name)

    cdef int _unpack_data(self) except -1:
        """
        Unpacks the packed data into a dictionary of Python values.
        """
        cdef DbObjectPickleBuffer buf
        buf = DbObjectPickleBuffer.__new__(DbObjectPickleBuffer)
        buf._populate_from_bytes(self.packed_data)
        buf.read_header(&self.image_flags, &self.image_version)
        self._unpack_data_from_buf(buf)
        self.packed_data = None

    cdef int _unpack_data_from_buf(self, DbObjectPickleBuffer buf) except -1:
        """
        Unpacks the data from the buffer into Python values.
        """
        cdef:
            dict unpacked_attrs = {}, unpacked_assoc_array = None
            ThinDbObjectTypeImpl typ_impl = self.type
            list unpacked_array = None
            ThinDbObjectAttrImpl attr
            uint32_t num_elements, i
            int32_t assoc_index
            object value
        if typ_impl.is_collection:
            if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                unpacked_assoc_array = {}
            else:
                unpacked_array = []
            buf.skip_raw_bytes(1)           # collection flags
            buf.read_length(&num_elements)
            for i in range(num_elements):
                if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                    buf.read_int32(&assoc_index)
                value = self._unpack_value(buf, typ_impl.element_dbtype,
                                           typ_impl.element_objtype)
                if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                    unpacked_assoc_array[assoc_index] = value
                else:
                    unpacked_array.append(value)
        else:
            unpacked_attrs = {}
            for attr in typ_impl.attrs:
                value = self._unpack_value(buf, attr.dbtype, attr.objtype)
                unpacked_attrs[attr.name] = value
        self.unpacked_attrs = unpacked_attrs
        self.unpacked_array = unpacked_array
        self.unpacked_assoc_array = unpacked_assoc_array

    cdef object _unpack_value(self, DbObjectPickleBuffer buf,
                              DbType dbtype, ThinDbObjectTypeImpl objtype):
        """
        Unpacks a single value and returns it.
        """
        cdef:
            uint8_t ora_type_num = dbtype._ora_type_num
            uint8_t csfrm = dbtype._csfrm
            ThinDbObjectImpl obj_impl
            ThinConnImpl conn_impl
            bint is_null
        if ora_type_num == TNS_DATA_TYPE_NUMBER:
            return buf.read_oracle_number(NUM_TYPE_FLOAT)
        elif ora_type_num == TNS_DATA_TYPE_BINARY_INTEGER:
            return buf.read_binary_integer()
        elif ora_type_num in (TNS_DATA_TYPE_VARCHAR, TNS_DATA_TYPE_CHAR):
            if csfrm == TNS_CS_NCHAR:
                conn_impl = self.type._conn_impl
                conn_impl._protocol._caps._check_ncharset_id()
            return buf.read_str(csfrm)
        elif ora_type_num == TNS_DATA_TYPE_RAW:
            return buf.read_bytes()
        elif ora_type_num == TNS_DATA_TYPE_BINARY_DOUBLE:
            return buf.read_binary_double()
        elif ora_type_num == TNS_DATA_TYPE_BINARY_FLOAT:
            return buf.read_binary_float()
        elif ora_type_num in (TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_TIMESTAMP,
                              TNS_DATA_TYPE_TIMESTAMP_LTZ,
                              TNS_DATA_TYPE_TIMESTAMP_TZ):
            return buf.read_date()
        elif ora_type_num in (TNS_DATA_TYPE_CLOB, TNS_DATA_TYPE_BLOB):
            conn_impl = self.type._conn_impl
            return buf.read_lob(conn_impl, dbtype)
        elif ora_type_num == TNS_DATA_TYPE_BOOLEAN:
            return buf.read_bool()
        elif ora_type_num == TNS_DATA_TYPE_INT_NAMED:
            buf.get_is_atomic_null(&is_null)
            if is_null:
                return None
            obj_impl = ThinDbObjectImpl.__new__(ThinDbObjectImpl)
            obj_impl.type = objtype
            if objtype.is_collection or self.type.is_collection:
                obj_impl.packed_data = buf.read_bytes()
            else:
                obj_impl._unpack_data_from_buf(buf)
            return PY_TYPE_DB_OBJECT._from_impl(obj_impl)
        errors._raise_err(errors.ERR_DB_TYPE_NOT_SUPPORTED, name=dbtype.name)

    def append_checked(self, object value):
        """
        Internal method for appending a value to a collection object.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            int32_t new_index
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            typ_impl = self.type
            if typ_impl.max_num_elements > 0 \
                    and len(self.unpacked_array) >= typ_impl.max_num_elements:
                errors._raise_err(errors.ERR_INVALID_COLL_INDEX_SET,
                                  index=len(self.unpacked_array),
                                  min_index=0,
                                  max_index=typ_impl.max_num_elements - 1)
            self.unpacked_array.append(value)
        else:
            self._ensure_assoc_keys()
            new_index = self.unpacked_assoc_keys[-1] + 1 \
                    if self.unpacked_assoc_keys else 0
            self.unpacked_assoc_array[new_index] = value
            self.unpacked_assoc_keys.append(new_index)

    def copy(self):
        """
        Internal method for creating a copy of an object.
        """
        cdef ThinDbObjectImpl copied_impl
        copied_impl = ThinDbObjectImpl.__new__(ThinDbObjectImpl)
        copied_impl.type = self.type
        copied_impl.flags = self.flags
        copied_impl.image_flags = self.image_flags
        copied_impl.image_version = self.image_version
        copied_impl.toid = self.toid
        copied_impl.packed_data = self.packed_data
        copied_impl.num_elements = self.num_elements
        if self.unpacked_attrs is not None:
            copied_impl.unpacked_attrs = self.unpacked_attrs.copy()
        if self.unpacked_array is not None:
            copied_impl.unpacked_array = list(self.unpacked_array)
        return copied_impl

    def delete_by_index(self, int32_t index):
        """
        Internal method for deleting an entry from a collection that is indexed
        by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            del self.unpacked_array[index]
        else:
            self.unpacked_assoc_keys = None
            del self.unpacked_assoc_array[index]

    def exists_by_index(self, int32_t index):
        """
        Internal method for determining if an entry exists in a collection that
        is indexed by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            return index >= 0 and index < len(self.unpacked_array)
        else:
            return index in self.unpacked_assoc_array

    def get_attr_value(self, ThinDbObjectAttrImpl attr):
        """
        Internal method for getting an attribute value.
        """
        self._ensure_unpacked()
        return self.unpacked_attrs[attr.name]

    def get_element_by_index(self, int32_t index):
        """
        Internal method for getting an entry from a collection that is indexed
        by integers.
        """
        self._ensure_unpacked()
        try:
            if self.unpacked_array is not None:
                return self.unpacked_array[index]
            else:
                return self.unpacked_assoc_array[index]
        except (KeyError, IndexError):
            errors._raise_err(errors.ERR_INVALID_COLL_INDEX_GET, index=index)

    def get_first_index(self):
        """
        Internal method for getting the first index from a collection that is
        indexed by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array:
            return 0
        elif self.unpacked_assoc_array:
            self._ensure_assoc_keys()
            return self.unpacked_assoc_keys[0]

    def get_last_index(self):
        """
        Internal method for getting the last index from a collection that is
        indexed by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array:
            return len(self.unpacked_array) - 1
        elif self.unpacked_assoc_array:
            self._ensure_assoc_keys()
            return self.unpacked_assoc_keys[-1]

    def get_next_index(self, int32_t index):
        """
        Internal method for getting the next index from a collection that is
        indexed by integers.
        """
        cdef int32_t i
        self._ensure_unpacked()
        if self.unpacked_array:
            if index + 1 < len(self.unpacked_array):
                return index + 1
        elif self.unpacked_assoc_array:
            self._ensure_assoc_keys()
            for i in self.unpacked_assoc_keys:
                if i > index:
                    return i

    def get_prev_index(self, int32_t index):
        """
        Internal method for getting the next index from a collection that is
        indexed by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array:
            if index > 0:
                return index - 1
        elif self.unpacked_assoc_array:
            self._ensure_assoc_keys()
            for i in reversed(self.unpacked_assoc_keys):
                if i < index:
                    return i

    def get_size(self):
        """
        Internal method for getting the size of a collection.
        """
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            return len(self.unpacked_array)
        else:
            return len(self.unpacked_assoc_array)

    def set_attr_value_checked(self, ThinDbObjectAttrImpl attr, object value):
        """
        Internal method for setting an attribute value.
        """
        self._ensure_unpacked()
        self.unpacked_attrs[attr.name] = value

    def set_element_by_index_checked(self, int32_t index, object value):
        """
        Internal method for setting an entry in a collection that is indexed by
        integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            try:
                self.unpacked_array[index] = value
            except IndexError:
                max_index = max(len(self.unpacked_array) - 1, 0)
                errors._raise_err(errors.ERR_INVALID_COLL_INDEX_SET,
                                  index=index, min_index=0,
                                  max_index=max_index)
        else:
            if index not in self.unpacked_assoc_array:
                self.unpacked_assoc_keys = None
            self.unpacked_assoc_array[index] = value

    def trim(self, int32_t num_to_trim):
        """
        Internal method for trimming a number of entries from a collection.
        """
        self._ensure_unpacked()
        if num_to_trim > 0:
            self.unpacked_array = self.unpacked_array[:-num_to_trim]


cdef class ThinDbObjectAttrImpl(BaseDbObjectAttrImpl):
    cdef:
        bytes oid


cdef class ThinDbObjectTypeImpl(BaseDbObjectTypeImpl):
    cdef:
        uint8_t collection_type, collection_flags, version
        uint32_t max_num_elements
        bint is_xml_type
        bytes oid

    def create_new_object(self):
        """
        Internal method for creating a new object.
        """
        cdef ThinDbObjectImpl obj_impl
        obj_impl = ThinDbObjectImpl.__new__(ThinDbObjectImpl)
        obj_impl.type = self
        obj_impl.toid = b'\x00\x22' + \
                bytes([TNS_OBJ_NON_NULL_OID, TNS_OBJ_HAS_EXTENT_OID]) + \
                self.oid + TNS_EXTENT_OID
        obj_impl.flags = TNS_OBJ_TOP_LEVEL
        obj_impl.image_flags = TNS_OBJ_IS_VERSION_81
        obj_impl.image_version = TNS_OBJ_IMAGE_VERSION
        obj_impl.unpacked_attrs = {}
        if self.is_collection:
            obj_impl.image_flags |= TNS_OBJ_IS_COLLECTION
            if self.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                obj_impl.unpacked_assoc_array = {}
            else:
                obj_impl.unpacked_array = []
        else:
            obj_impl.image_flags |= TNS_OBJ_NO_PREFIX_SEG
            for attr in self.attrs:
                obj_impl.unpacked_attrs[attr.name] = None
        return obj_impl


cdef class ThinDbObjectTypeSuperCache:
    cdef:
        dict caches
        object lock
        int cache_num

    def __init__(self):
        self.caches = {}
        self.cache_num = 0
        self.lock = threading.Lock()


cdef class ThinDbObjectTypeCache:
    cdef:
        object return_value_var, full_name_var, oid_var, tds_var
        object meta_cursor, attrs_ref_cursor_var, version_var
        object schema_var, package_name_var, name_var
        ThinConnImpl conn_impl
        dict types_by_oid
        dict types_by_name
        list partial_types

    cdef int _clear_meta_cursor(self) except -1:
        """
        Clears the cursor used for searching metadata. This is needed when
        returning a connection to the pool since user-level objects are
        retained.
        """
        if self.meta_cursor is not None:
            self.meta_cursor.close()
            self.meta_cursor = None
            self.return_value_var = None
            self.full_name_var = None
            self.oid_var = None
            self.tds_var = None
            self.attrs_ref_cursor_var = None
            self.version_var = None
            self.schema_var = None
            self.package_name_var = None
            self.name_var = None

    cdef int _determine_element_type_csfrm(self, ThinDbObjectTypeImpl typ_impl,
                                           uint8_t* csfrm) except -1:
        """
        Determine the element type's character set form. This is only needed
        for CLOB and NCLOB where this information is not stored in the TDS.
        """
        cdef:
            object cursor
            str type_name
        cursor = self.meta_cursor.connection.cursor()
        if typ_impl.package_name is not None:
            cursor.execute("""
                    select elem_type_name
                    from all_plsql_coll_types
                    where owner = :owner
                      and package_name = :package_name
                      and type_name = :name""",
                    owner=typ_impl.schema,
                    package_name=typ_impl.package_name,
                    name=typ_impl.name)
        else:
            cursor.execute("""
                    select elem_type_name
                    from all_coll_types
                    where owner = :owner
                      and type_name = :name""",
                    owner=typ_impl.schema,
                    name=typ_impl.name)

        type_name, = cursor.fetchone()
        if type_name == "NCLOB":
            csfrm[0] = TNS_CS_NCHAR
        else:
            csfrm[0] = TNS_CS_IMPLICIT

    cdef int _determine_element_objtype(self,
                                        ThinDbObjectTypeImpl impl) except -1:
        """
        Determine the element type's object type. This is needed when
        processing collections with object as the element type since this
        information is not available in the TDS.
        """
        cdef:
            str schema, name, package_name = None
            object cursor
        cursor = self.meta_cursor.connection.cursor()
        if impl.package_name is not None:
            cursor.execute("""
                    select
                        elem_type_owner,
                        elem_type_package,
                        elem_type_name
                    from all_plsql_coll_types
                    where owner = :owner
                      and package_name = :package_name
                      and type_name = :name""",
                    owner=impl.schema,
                    package_name=impl.package_name,
                    name=impl.name)
            schema, package_name, name = cursor.fetchone()
        else:
            cursor.execute("""
                    select
                        elem_type_owner,
                        elem_type_name
                    from all_coll_types
                    where owner = :owner
                      and type_name = :name""",
                    owner=impl.schema,
                    name=impl.name)
            schema, name = cursor.fetchone()
        impl.element_objtype = self.get_type_for_info(None, schema,
                                                      package_name, name)

    cdef int _initialize(self, ThinConnImpl conn_impl) except -1:
        self.types_by_oid = {}
        self.types_by_name = {}
        self.partial_types = []
        self.conn_impl = conn_impl

    cdef int _init_meta_cursor(self, object conn) except -1:
        """
        Initializes the cursor that fetches the type metadata.
        """
        cursor = conn.cursor()
        self.return_value_var = cursor.var(DB_TYPE_BINARY_INTEGER)
        self.tds_var = cursor.var(bytes)
        self.full_name_var = cursor.var(str)
        self.schema_var = cursor.var(str)
        self.package_name_var = cursor.var(str)
        self.name_var = cursor.var(str)
        self.oid_var = cursor.var(bytes)
        self.version_var = cursor.var(DB_TYPE_BINARY_INTEGER)
        self.attrs_ref_cursor_var = cursor.var(DB_TYPE_CURSOR)
        cursor.setinputsizes(ret_val=self.return_value_var,
                             tds=self.tds_var,
                             full_name=self.full_name_var,
                             oid=self.oid_var,
                             schema=self.schema_var,
                             package_name=self.package_name_var,
                             name=self.name_var,
                             version=self.version_var,
                             attrs_rc=self.attrs_ref_cursor_var)
        cursor.prepare("""
            declare
                t_Instantiable              varchar2(3);
                t_SuperTypeOwner            varchar2(128);
                t_SuperTypeName             varchar2(128);
                t_SubTypeRefCursor          sys_refcursor;
                t_Pos                       pls_integer;
            begin
                :ret_val := dbms_pickler.get_type_shape(:full_name, :oid,
                    :version, :tds, t_Instantiable, t_SuperTypeOwner,
                    t_SuperTypeName, :attrs_rc, t_SubTypeRefCursor);
                :package_name := null;
                if substr(:full_name, length(:full_name) - 7) = '%ROWTYPE' then
                    t_Pos := instr(:full_name, '.');
                    :schema := substr(:full_name, 1, t_Pos - 1);
                    :name := substr(:full_name, t_Pos + 1);
                else
                    begin
                        select owner, type_name
                        into :schema, :name
                        from all_types
                        where type_oid = :oid;
                    exception
                    when no_data_found then
                        begin
                            select owner, package_name, type_name
                            into :schema, :package_name, :name
                            from all_plsql_types
                            where type_oid = :oid;
                        exception
                        when no_data_found then
                            null;
                        end;
                    end;
                end if;
            end;""")
        self.meta_cursor = cursor

    cdef int _parse_element_type(self, ThinDbObjectTypeImpl typ_impl,
                                 TDSBuffer buf) except -1:
        """
        Parses the element type from the TDS buffer.
        """
        cdef uint8_t attr_type, ora_type_num = 0, csfrm = 0
        buf.read_ub1(&attr_type)
        if attr_type in (TNS_OBJ_TDS_TYPE_NUMBER, TNS_OBJ_TDS_TYPE_FLOAT):
            ora_type_num = TNS_DATA_TYPE_NUMBER
        elif attr_type in (TNS_OBJ_TDS_TYPE_VARCHAR, TNS_OBJ_TDS_TYPE_CHAR):
            buf.skip_raw_bytes(2)           # maximum length
            buf.read_ub1(&csfrm)
            csfrm = csfrm & 0x7f
            if attr_type == TNS_OBJ_TDS_TYPE_VARCHAR:
                ora_type_num = TNS_DATA_TYPE_VARCHAR
            else:
                ora_type_num = TNS_DATA_TYPE_CHAR
        elif attr_type == TNS_OBJ_TDS_TYPE_RAW:
            ora_type_num = TNS_DATA_TYPE_RAW
        elif attr_type == TNS_OBJ_TDS_TYPE_BINARY_FLOAT:
            ora_type_num = TNS_DATA_TYPE_BINARY_FLOAT
        elif attr_type == TNS_OBJ_TDS_TYPE_BINARY_DOUBLE:
            ora_type_num = TNS_DATA_TYPE_BINARY_DOUBLE
        elif attr_type == TNS_OBJ_TDS_TYPE_DATE:
            ora_type_num = TNS_DATA_TYPE_DATE
        elif attr_type == TNS_OBJ_TDS_TYPE_TIMESTAMP:
            ora_type_num = TNS_DATA_TYPE_TIMESTAMP
        elif attr_type == TNS_OBJ_TDS_TYPE_TIMESTAMP_LTZ:
            ora_type_num = TNS_DATA_TYPE_TIMESTAMP_LTZ
        elif attr_type == TNS_OBJ_TDS_TYPE_TIMESTAMP_TZ:
            ora_type_num = TNS_DATA_TYPE_TIMESTAMP_TZ
        elif attr_type == TNS_OBJ_TDS_TYPE_BOOLEAN:
            ora_type_num = TNS_DATA_TYPE_BOOLEAN
        elif attr_type == TNS_OBJ_TDS_TYPE_CLOB:
            ora_type_num = TNS_DATA_TYPE_CLOB
            self._determine_element_type_csfrm(typ_impl, &csfrm)
        elif attr_type == TNS_OBJ_TDS_TYPE_BLOB:
            ora_type_num = TNS_DATA_TYPE_BLOB
        elif attr_type == TNS_OBJ_TDS_TYPE_OBJ:
            ora_type_num = TNS_DATA_TYPE_INT_NAMED
            self._determine_element_objtype(typ_impl)
        else:
            errors._raise_err(errors.ERR_TDS_TYPE_NOT_SUPPORTED, num=attr_type)
        typ_impl.element_dbtype = DbType._from_ora_type_and_csfrm(ora_type_num,
                                                                  csfrm)

    cdef int _parse_tds(self, ThinDbObjectTypeImpl typ_impl,
                        bytes tds) except -1:
        """
        Parses the TDS for the type. This is only needed for collection types,
        so if the TDS is determined to be for an object type, the remaining
        information is skipped.
        """
        cdef:
            uint32_t element_pos
            uint16_t num_attrs
            uint8_t attr_type
            TDSBuffer buf

        # parse initial TDS bytes
        buf = TDSBuffer.__new__(TDSBuffer)
        buf._populate_from_bytes(tds)
        buf.skip_raw_bytes(4)               # end offset
        buf.skip_raw_bytes(2)               # version op code and version
        buf.skip_raw_bytes(2)               # unknown

        # if the number of attributes exceeds 1, the type cannot refer to a
        # collection, so nothing further needs to be done
        buf.read_uint16(&num_attrs)
        if num_attrs > 1:
            return 0

        # continue parsing TDS bytes to discover if type refers to a collection
        buf.skip_raw_bytes(1)               # TDS attributes?
        buf.skip_raw_bytes(1)               # start ADT op code
        buf.skip_raw_bytes(2)               # ADT number (always zero)
        buf.skip_raw_bytes(4)               # offset to index table

        # if type of first attribute is not a collection, nothing further needs
        # to be done
        buf.read_ub1(&attr_type)
        if attr_type != TNS_OBJ_TDS_TYPE_COLL:
            return 0
        typ_impl.is_collection = True

        # continue parsing TDS to determine element type
        buf.read_uint32(&element_pos)
        buf.read_uint32(&typ_impl.max_num_elements)
        buf.read_ub1(&typ_impl.collection_type)
        if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
            typ_impl.collection_flags = TNS_OBJ_HAS_INDEXES
        buf.skip_to(element_pos)
        self._parse_element_type(typ_impl, buf)

    cdef int _populate_type_info(self, str name,
                                 ThinDbObjectTypeImpl typ_impl) except -1:
        """
        Populate the type information given the name of the type.
        """
        cdef:
            ThinDbObjectAttrImpl attr_impl
            ssize_t pos, name_length
            list name_components
            object attrs_rc
        self.full_name_var.setvalue(0, name)
        self.meta_cursor.execute(None)
        if self.return_value_var.getvalue() != 0:
            errors._raise_err(errors.ERR_INVALID_OBJECT_TYPE_NAME, name=name)
        typ_impl.version = self.version_var.getvalue()
        if typ_impl.oid is None:
            typ_impl.oid = self.oid_var.getvalue()
            self.types_by_oid[typ_impl.oid] = typ_impl
        if typ_impl.schema is None:
            typ_impl.schema = self.schema_var.getvalue()
            typ_impl.package_name = self.package_name_var.getvalue()
            typ_impl.name = self.name_var.getvalue()
            typ_impl.is_xml_type = \
                    (typ_impl.schema == "SYS" and typ_impl.name == "XMLTYPE")
        self._parse_tds(typ_impl, self.tds_var.getvalue())
        typ_impl.attrs = []
        typ_impl.attrs_by_name = {}
        attrs_rc = self.attrs_ref_cursor_var.getvalue()
        for cursor_version, attr_name, attr_num, attr_type_name, \
                attr_type_owner, attr_type_package, attr_type_oid, \
                attr_instantiable, attr_super_type_owner, \
                attr_super_type_name in attrs_rc:
            attr_impl = ThinDbObjectAttrImpl.__new__(ThinDbObjectAttrImpl)
            attr_impl.name = attr_name
            if attr_type_owner is not None:
                attr_impl.dbtype = DB_TYPE_OBJECT
                attr_impl.objtype = self.get_type_for_info(attr_type_oid,
                                                           attr_type_owner,
                                                           attr_type_package,
                                                           attr_type_name)
            else:
                attr_impl.dbtype = DbType._from_ora_name(attr_type_name)
            typ_impl.attrs.append(attr_impl)
            typ_impl.attrs_by_name[attr_name] = attr_impl

    cdef ThinDbObjectTypeImpl get_type(self, object conn, str name):
        """
        Returns the database object type given its name. The cache is first
        searched and if it is not found, the database is searched and the
        result stored in the cache.
        """
        cdef ThinDbObjectTypeImpl typ_impl
        typ_impl = self.types_by_name.get(name)
        if typ_impl is None:
            if self.meta_cursor is None:
                self._init_meta_cursor(conn)
            typ_impl = ThinDbObjectTypeImpl.__new__(ThinDbObjectTypeImpl)
            typ_impl._conn_impl = self.conn_impl
            self._populate_type_info(name, typ_impl)
            self.types_by_oid[typ_impl.oid] = typ_impl
            self.types_by_name[name] = typ_impl
            self.populate_partial_types(conn)
        return typ_impl

    cdef ThinDbObjectTypeImpl get_type_for_info(self, bytes oid, str schema,
                                                str package_name, str name):
        """
        Returns a type for the specified fetch info, if one has already been
        cached. If not, a new type object is created and cached. It is also
        added to the partial_types list which will be fully populated once the
        current execute has completed.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            str full_name
        if package_name is not None:
            full_name = f"{schema}.{package_name}.{name}"
        else:
            full_name = f"{schema}.{name}"
        if oid is not None:
            typ_impl = self.types_by_oid.get(oid)
        else:
            typ_impl = self.types_by_name.get(full_name)
        if typ_impl is None:
            typ_impl = ThinDbObjectTypeImpl.__new__(ThinDbObjectTypeImpl)
            typ_impl._conn_impl = self.conn_impl
            typ_impl.oid = oid
            typ_impl.schema = schema
            typ_impl.package_name = package_name
            typ_impl.name = name
            typ_impl.is_xml_type = (schema == "SYS" and name == "XMLTYPE")
            if oid is not None:
                self.types_by_oid[oid] = typ_impl
            self.types_by_name[full_name] = typ_impl
            self.partial_types.append(typ_impl)
        return typ_impl

    cdef int populate_partial_types(self, object conn) except -1:
        """
        Populate any partial types that were discovered earlier. Since
        populating an object type might result in additional object types being
        discovered, object types are popped from the partial types list until
        the list is empty.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            str full_name, name, suffix
        while self.partial_types:
            typ_impl = self.partial_types.pop()
            if self.meta_cursor is None:
                self._init_meta_cursor(conn)
            suffix = "%ROWTYPE"
            if typ_impl.name.endswith(suffix):
                name = typ_impl.name[:-len(suffix)]
            else:
                name = typ_impl.name
                suffix = ""
            if typ_impl.package_name is not None:
                full_name = f'"{typ_impl.schema}".' + \
                            f'"{typ_impl.package_name}".' + \
                            f'"{name}"{suffix}'
            else:
                full_name = f'"{typ_impl.schema}"."{name}"{suffix}'
            self._populate_type_info(full_name, typ_impl)


# global cache of database object types
# since the database object types require a reference to the connection (in
# order to be able to manage LOBs), storing the cache on the connection would
# involve creating a circular reference
cdef ThinDbObjectTypeSuperCache DB_OBJECT_TYPE_SUPER_CACHE = \
        ThinDbObjectTypeSuperCache()


cdef int create_new_dbobject_type_cache(ThinConnImpl conn_impl) except -1:
    """
    Creates a new database object type cache and returns its identifier.
    """
    cdef:
        ThinDbObjectTypeCache cache
        int cache_num
    with DB_OBJECT_TYPE_SUPER_CACHE.lock:
        DB_OBJECT_TYPE_SUPER_CACHE.cache_num += 1
        cache_num = DB_OBJECT_TYPE_SUPER_CACHE.cache_num
    cache = ThinDbObjectTypeCache.__new__(ThinDbObjectTypeCache)
    cache._initialize(conn_impl)
    DB_OBJECT_TYPE_SUPER_CACHE.caches[cache_num] = cache
    return cache_num


cdef ThinDbObjectTypeCache get_dbobject_type_cache(int cache_num):
    """
    Returns the database object type cache given its identifier.
    """
    return DB_OBJECT_TYPE_SUPER_CACHE.caches[cache_num]


cdef int remove_dbobject_type_cache(int cache_num) except -1:
    """
    Removes the sub cache given its identifier.
    """
    del DB_OBJECT_TYPE_SUPER_CACHE.caches[cache_num]
